# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cSpell: ignore vsize

import abc
from typing import List, Optional, Any, Iterator
from ansys.scadeone.core.common.exception import ScadeOneException
import ansys.scadeone.core.svc.simdata.core as core


###########################################################################
# data types


class Type:
    """Represents an abstract data type."""

    def __init__(self, type_id: int, name: str = "") -> None:
        self._type_id = type_id
        self._name = name

    @property
    def type_id(self) -> int:
        return self._type_id

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name


PredefinedTypeKind = core.PredefinedType


class PredefinedType(Type):
    """
    Predefined Swan types are not stored in file.
    They have hard-coded identifiers that can be used in user types definitions.
    """

    _instances = {}

    def __init__(self, type_id: int) -> None:
        try:
            self._kind = PredefinedTypeKind(type_id)
        except:  # noqa: E722
            raise ScadeOneException("cannot create predefined type with id {0}".format(type_id))
        super().__init__(type_id, self.__str__())
        PredefinedType._instances[type_id] = self

    def __str__(self) -> str:
        if self._kind == PredefinedTypeKind.CHAR:
            return "char"
        elif self._kind == PredefinedTypeKind.BOOL:
            return "bool"
        elif self._kind == PredefinedTypeKind.INT8:
            return "int8"
        elif self._kind == PredefinedTypeKind.INT16:
            return "int16"
        elif self._kind == PredefinedTypeKind.INT32:
            return "int32"
        elif self._kind == PredefinedTypeKind.INT64:
            return "int64"
        elif self._kind == PredefinedTypeKind.UINT8:
            return "uint8"
        elif self._kind == PredefinedTypeKind.UINT16:
            return "uint16"
        elif self._kind == PredefinedTypeKind.UINT32:
            return "uint32"
        elif self._kind == PredefinedTypeKind.UINT64:
            return "uint64"
        elif self._kind == PredefinedTypeKind.FLOAT32:
            return "float32"
        elif self._kind == PredefinedTypeKind.FLOAT64:
            return "float64"
        return "<unknown>"

    @property
    def kind(self):
        return self._kind

    @classmethod
    def get(cls, type_id: int) -> Optional["PredefinedType"]:
        return cls._instances[type_id] if type_id in cls._instances else None


Char = PredefinedType(int(PredefinedTypeKind.CHAR))
Bool = PredefinedType(int(PredefinedTypeKind.BOOL))
Int8 = PredefinedType(int(PredefinedTypeKind.INT8))
Int16 = PredefinedType(int(PredefinedTypeKind.INT16))
Int32 = PredefinedType(int(PredefinedTypeKind.INT32))
Int64 = PredefinedType(int(PredefinedTypeKind.INT64))
UInt8 = PredefinedType(int(PredefinedTypeKind.UINT8))
UInt16 = PredefinedType(int(PredefinedTypeKind.UINT16))
UInt32 = PredefinedType(int(PredefinedTypeKind.UINT32))
UInt64 = PredefinedType(int(PredefinedTypeKind.UINT64))
Float32 = PredefinedType(int(PredefinedTypeKind.FLOAT32))
Float64 = PredefinedType(int(PredefinedTypeKind.FLOAT64))


class StructTypeField:
    """Structure type's fields"""

    def __init__(self, name: str, offset: int, sd_type: Type) -> None:
        self._name = name
        self._offset = offset
        self._sd_type = sd_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def offset(self) -> int:
        return self._offset

    @property
    def sd_type(self) -> Type:
        return self._sd_type


class StructType(Type):
    """Structure type defined with structure type fields"""

    def __init__(self, type_id: int, fields: List[StructTypeField], name: str = "") -> None:
        self._fields = fields
        super().__init__(type_id, name)

    @property
    def fields(self) -> List[StructTypeField]:
        return self._fields

    def __str__(self) -> str:
        return "struct{" + ",".join([f.name + ":" + str(f.sd_type) for f in self.fields]) + "}"


class ArrayType(Type):
    """Multidimensional array type"""

    def __init__(self, type_id: int, base_type: Type, dims: List[int], name: str = "") -> None:
        self._base_type = base_type
        self._dims = dims
        super().__init__(type_id, name)

    @property
    def base_type(self) -> Type:
        return self._base_type

    @property
    def dims(self) -> List[int]:
        return self._dims

    def __str__(self) -> str:
        return str(self.base_type) + "^" + "^".join([str(d) for d in self.dims[::-1]])


class EnumTypeValue:
    """Enumeration type value"""

    def __init__(self, name: str, int_value: int) -> None:
        self._name = name
        self._int_value = int_value

    @property
    def name(self) -> str:
        return self._name

    @property
    def int_value(self) -> int:
        return self._int_value


class EnumType(Type):
    """Enumeration type defined with enumeration values"""

    def __init__(
        self, type_id: int, base_type: PredefinedType, values: List[EnumTypeValue], name: str = ""
    ) -> None:
        self._base_type = base_type
        self._values = values
        super().__init__(type_id, name)

    @property
    def base_type(self) -> PredefinedType:
        return self._base_type

    @property
    def values(self) -> List[EnumTypeValue]:
        return self._values

    def __str__(self) -> str:
        return (
            self.name
            + " enum{"
            + ",".join([v.name + ":" + str(v.int_value) for v in self.values])
            + "}"
        )


class VariantTypeConstructor:
    """Variant type constructor"""

    def __init__(self, name: str, value_type: Optional[Type]) -> None:
        self._name = name
        self._value_type = value_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def value_type(self) -> Optional[Type]:
        return self._value_type


class VariantType(Type):
    """Variant type defined with variant type constructors"""

    def __init__(
        self, type_id: int, constructors: List[VariantTypeConstructor], name: str = ""
    ) -> None:
        self._constructors = constructors
        super().__init__(type_id, name)

    @property
    def constructors(self) -> List[VariantTypeConstructor]:
        return self._constructors

    def find_constructor(self, name: str) -> Optional[VariantTypeConstructor]:
        for c in self._constructors:
            if c.name == name:
                return c
        return None

    def __str__(self) -> str:
        return "|".join(
            [
                c.name + "{" + (str(c.value_type) if c.value_type is not None else "") + "}"
                for c in self.constructors
            ]
        )


PfnVsizeGetBytesSize = core.sd_pfn_vsize_get_bytes_size_t

PfnVsizeToBytes = core.sd_pfn_vsize_to_bytes_t


class ImportedType(Type):
    """Imported Types (stored as byte arrays)"""

    def __init__(
        self,
        type_id: int,
        mem_size: int,
        vsize: bool = False,
        pfn_vsize_get_bytes_size: PfnVsizeGetBytesSize = None,
        pfn_vsize_to_bytes: PfnVsizeToBytes = None,
        name: str = "",
    ) -> None:
        self._mem_size = mem_size
        self._vsize = vsize
        self._pfn_vsize_get_bytes_size = pfn_vsize_get_bytes_size
        self._pfn_vsize_to_bytes = pfn_vsize_to_bytes
        super().__init__(type_id, name)

    @property
    def mem_size(self) -> int:
        return self._mem_size

    @property
    def vsize(self) -> bool:
        return self._vsize

    @property
    def pfn_vsize_get_bytes_size(self) -> PfnVsizeGetBytesSize:
        return self._pfn_vsize_get_bytes_size

    @pfn_vsize_get_bytes_size.setter
    def pfn_vsize_get_bytes_size(self, pfn_vsize_get_bytes_size: PfnVsizeGetBytesSize):
        self._pfn_vsize_get_bytes_size = pfn_vsize_get_bytes_size

    @property
    def pfn_vsize_to_bytes(self) -> PfnVsizeToBytes:
        return self._pfn_vsize_to_bytes

    @pfn_vsize_to_bytes.setter
    def pfn_vsize_to_bytes(self, pfn_vsize_to_bytes: PfnVsizeToBytes):
        self._pfn_vsize_to_bytes = pfn_vsize_to_bytes

    def __str__(self) -> str:
        return "<variable size imported>" if self.vsize else "<imported>"


class Value:
    """Interface for all element types values that shall provide a readable string representation"""

    pass


class PredefinedValue(Value):
    """Values for predefined Swan Types"""

    pass


class NoneValue(Value):
    """None Values"""

    def __str__(self):
        return "none"


class PredefinedCharValue(PredefinedValue):
    """Values for predefined type char"""

    def __init__(self, value: core.sd_uint8_t) -> None:
        self._value = value

    def __str__(self):
        return "'" + chr(self._value.value) + "'"


class PredefinedBoolValue(PredefinedValue):
    """Values for predefined type boolean"""

    def __init__(self, value: core.sd_uint8_t) -> None:
        self._value = True if value else False

    def __str__(self):
        return "true" if self._value else "false"


class PredefinedInt8Value(PredefinedValue):
    """Values for predefined type int 8"""

    def __init__(self, value: core.sd_int8_t) -> None:
        self._value = value

    def __str__(self):
        return str(self._value.value)


class PredefinedInt16Value(PredefinedValue):
    """Values for predefined type int 16"""

    def __init__(self, value: core.sd_int16_t) -> None:
        self._value = value

    def __str__(self):
        return str(self._value.value)


class PredefinedInt32Value(PredefinedValue):
    """Values for predefined type int 32"""

    def __init__(self, value: core.sd_int32_t) -> None:
        self._value = value

    def __str__(self):
        return str(self._value.value)


class PredefinedInt64Value(PredefinedValue):
    """Values for predefined type int 64"""

    def __init__(self, value: core.sd_int64_t) -> None:
        self._value = value

    def __str__(self):
        return str(self._value.value)


class PredefinedUInt8Value(PredefinedValue):
    """Values for predefined type unsigned int 8"""

    def __init__(self, value: core.sd_uint8_t) -> None:
        self._value = value

    def __str__(self):
        return str(self._value.value)


class PredefinedUInt16Value(PredefinedValue):
    """Values for predefined type unsigned int 16"""

    def __init__(self, value: core.sd_uint16_t) -> None:
        self._value = value

    def __str__(self):
        return str(self._value.value)


class PredefinedUInt32Value(PredefinedValue):
    """Values for predefined type unsigned int 32"""

    def __init__(self, value: core.sd_uint32_t) -> None:
        self._value = value

    def __str__(self):
        return str(self._value.value)


class PredefinedUInt64Value(PredefinedValue):
    """Values for predefined type unsigned int 64"""

    def __init__(self, value: core.sd_uint64_t) -> None:
        self._value = value

    def __str__(self):
        return str(self._value.value)


class PredefinedFloat32Value(PredefinedValue):
    """Values for predefined type float 32"""

    def __init__(self, value: core.sd_float32_t) -> None:
        self._value = value

    def __str__(self):
        return f"{self._value.value:.5g}"


class PredefinedFloat64Value(PredefinedValue):
    """Values for predefined type float 64"""

    def __init__(self, value: core.sd_float64_t) -> None:
        self._value = value

    def __str__(self):
        return f"{self._value.value:.5g}"


class ListValue(Value):
    """Values for list types"""

    def __init__(self, values: List[Value]) -> None:
        self._values = values

    def __str__(self):
        return "(" + ",".join([str(v) for v in self._values]) + ")"


class EnumValue(Value):
    """Values for enumeration types"""

    def __init__(self, name: str) -> None:
        self._name = name

    def __str__(self):
        return self._name


class VariantValue(Value):
    """Values for variant types"""

    def __init__(self, name: str, value: Optional[Value]) -> None:
        self._name = name
        self._value = value

    def __str__(self):
        return self._name + "{" + (str(self._value) if self._value is not None else "") + "}"


class UntypedVariantConstructorValue(Value):
    """No value to return for untyped variant constructor"""

    def __str__(self):
        return ""


class ImportedValue(Value):
    """Values for imported types"""

    def __init__(self, bytes_data) -> None:
        self._bytes_data = bytes_data

    def __str__(self):
        return (
            "" if self._bytes_data is None else "".join([f"{by:0>4X}" for by in self._bytes_data])
        )


ElementKind = core.SdeKind


class ElementBase(metaclass=abc.ABCMeta):
    def __init__(
        self,
        file: "FileBase",
        parent: Optional["ElementBase"],
        name: str,
        sd_type: Optional[Type],
        kind: ElementKind,
    ) -> None:
        self._file = file
        self._parent = parent
        self._name = name
        self._sd_type = sd_type
        self._kind = kind
        self._children_elements = []  # type: List[ElementBase]

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    @abc.abstractmethod
    def name(self, name: str) -> None:
        pass

    @property
    def sd_type(self) -> Optional[Type]:
        return self._sd_type

    @sd_type.setter
    @abc.abstractmethod
    def sd_type(self, sd_type: Optional[Type]) -> None:
        pass

    @property
    def kind(self) -> ElementKind:
        return self._kind

    @kind.setter
    @abc.abstractmethod
    def kind(self, kind: ElementKind) -> None:
        pass

    @property
    def parent(self) -> Optional["ElementBase"]:
        return self._parent

    @property
    def children_elements(self) -> List["ElementBase"]:
        return self._children_elements

    @property
    def nb_parents(self):
        if self.parent is None:
            return 0
        return self.parent.nb_parents + 1

    @abc.abstractmethod
    def add_child_element(
        self, name: str, sd_type: Type = None, kind: ElementKind = ElementKind.NONE
    ) -> "ElementBase":
        pass

    @abc.abstractmethod
    def remove_child_element(self, element: "ElementBase") -> None:
        pass

    def find_child_element(self, name: str) -> Optional["ElementBase"]:
        for e in self.children_elements:
            if e.name == name:
                return e
        return None

    @abc.abstractmethod
    def append_value(self, py_value: Any) -> None:
        pass

    def append_nones(self, count) -> None:
        pass

    def append_values(self, py_values: List[Any], repeat_factor: Optional[int] = 1) -> None:
        pass

    @abc.abstractmethod
    def read_values(self, start: Optional[int] = None, n: Optional[int] = None) -> Iterator[Value]:
        pass

    @abc.abstractmethod
    def clear_values(self) -> None:
        pass

    def __str__(self) -> str:
        indent = "    " * self.nb_parents
        type_str = "" if self.sd_type is None else ": " + str(self.sd_type)
        kind_str = self.kind.name if self.kind != ElementKind.NONE else ""
        s = "{0}{1} {2}{3}: ".format(indent, kind_str, self.name, type_str)
        for v in self.read_values():
            s += str(v) + " | "
        s += "\n"
        for child in self.children_elements:
            s += str(child)
        return s


class FileBase:
    def __init__(self) -> None:
        self._elements = []  # type: List[ElementBase]

    @property
    def elements(self) -> List[ElementBase]:
        return self._elements

    @abc.abstractmethod
    def add_element(
        self, name: str, sd_type: Type = None, kind: ElementKind = ElementKind.NONE
    ) -> ElementBase:
        pass

    @abc.abstractmethod
    def remove_element(self, element: ElementBase) -> None:
        pass

    @abc.abstractmethod
    def get_version(self) -> str:
        pass

    @abc.abstractmethod
    def close(self) -> None:
        pass

    def find_element(self, name: str) -> Optional[ElementBase]:
        for e in self.elements:
            if e.name == name:
                return e
        return None

    def __str__(self) -> str:
        s = "*** Elements:\n"
        for elem in self.elements:
            s += str(elem)
        return s



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains classes that are used by all the other
language-related modules.

It contains base classes for various language constructs.
"""

from abc import ABC
from collections import namedtuple
from collections.abc import Iterable as abcIterable
from enum import Enum, auto
import re
from typing import Iterable, List, Optional, Tuple, Union

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.interfaces import IProject
from ansys.scadeone.core.swan.pragmas import PragmaBase, Pragma

Owner = Union["SwanItem", IProject, None]


class SwanItem(ABC):  # numpydoc ignore=PR01
    """Base class for Scade objects."""

    def __init__(self) -> None:
        self._owner = None
        super().__init__()

    @property
    def owner(self) -> Owner:
        """Owner of current Swan construct."""
        return self._owner

    @owner.setter
    def owner(self, owner: Owner) -> None:  # numpydoc ignore=PR01
        """Set the owner of the Swan construct."""
        self._owner = owner

    @staticmethod
    def set_owner(owner: Owner, children: Union["SwanItem", Iterable["SwanItem"]]) -> None:
        """Helper to set *owner* as the owner of each item in the Iterable *items*.

        Parameters
        ----------
        owner : SwanItem
            Owner of the items.
        children : Union[SwanItem, Iterable[SwanItem]]
            Items to set owner.
        """
        if isinstance(children, abcIterable):
            for item in children:
                item.owner = owner
            return
        if children is not None:
            children.owner = owner

    def get_full_path(self) -> str:
        """Full path of Swan construct."""
        raise ScadeOneException(f"SwanItem.get_full_path(): not implemented for {type(self)}")

    def __str__(self) -> str:
        raise Exception(f"__str__ not implemented for {type(self)}")

    @property
    def is_protected(self) -> bool:
        """Tell if a construct item is syntactically protected with some markup
        and is stored as a string (without the markup)."""
        return False

    @property
    def module(self) -> Union["ModuleBase", None]:
        """Module containing the item.

        Returns:
            ModuleBase: module container, see :py:class:`ModuleBody`
            and :py:class:`ModuleInterface` or None if the object is itself a module.
        """
        if isinstance(self, ModuleBase):
            return None
        if isinstance(self.owner, ModuleBase):
            return self.owner
        if self.owner is not None:
            return self.owner.module
        raise ScadeOneException("owner property is None")

    @property
    def model(self) -> "Model":  # noqa: F821
        """Return model containing the Swan item."""
        if isinstance(self, ModuleBase):
            module = self
        else:
            module = self.module
        model = module.owner
        if model is None:
            raise ScadeOneException("Module owner not found")
        return model


class ModuleBase(SwanItem):  # numpydoc ignore=PR01
    """Base class for modules."""

    def __init__(self) -> None:
        super().__init__()

    def get_use_directive(self, name: str) -> Union["UseDirective", None]:  # noqa: F821
        assert False


# ========================================================
# Miscellaneous general definitions
# Type for parsing an integer from a string.
# See class Numeric RE
IntegerTuple = namedtuple(
    "IntegerTuple",
    ["value", "is_bin", "is_oct", "is_hex", "is_dec", "is_signed", "size"],
)

# for parsing a float from a string
# see class Numeric RE
FloatTuple = namedtuple("FloatTuple", ["value", "mantissa", "exp", "size"])


class SwanRE:  # numpydoc ignore=PR01
    """Container of compiled regular expressions. These expressions can be matched with
    some strings. Some regular expressions use groups to extract parts.

    Attributes:

    TypedInteger: regular expressions for integers with post type (_i, _ui).
             Integer part is in the *value* group, type in the *type* group,
             and size in the *size* group.

    TypeFloat: regular expression for floats with post type (_f).
           Mantissa is in the *mantissa* group, exponent in the *exp* group,
           type in the *type* group, and size in the *size* group.

    """

    def __init__(self) -> None:
        # Hide default init with kwargs and args (introspection)
        pass

    TypedInteger = re.compile(
        r"""^
    (?P<value>                    # value group
     (?:0b[01]+)                  # binary
    |(?:0o[0-7]+)                 # octal
    |(?:0x[0-9a-fA-F]+)           # hexadecimal
    |(?:\d+))                     # decimal
    (?:(?P<type>_ui|_i)(?P<size>8|16|32|64))? # type group & size
    $""",
        re.VERBOSE,
    )

    TypedFloat = re.compile(
        r"""^
    (?P<value>                              # float part
    (?P<mantissa>(?:\d+\.\d*)|(?:\d*\.\d+))  # mantissa
    (?:[eE](?P<exp>[+-]?\d+))?)              # exponent
    (?:_f(?P<size>32|64))?                   # size
    $""",
        re.VERBOSE,
    )

    @classmethod
    def parse_integer(cls, string: str, minus: bool = False) -> Union[IntegerTuple, None]:
        """Matches a string representing an integer and returns
        a description of that integer as an IntegerTuple.

        Parameters
        ----------
        string : str
            String representing an integer, with or without type information.
        minus : bool
            True when the value is preceded with a '-' minus operator.

        Returns
        -------
        IntegerTuple or None
            If the string value matches SwanRE.TypedInteger pattern, an
            IntegerTuple is returned. It is a namedtuple with fields:

            - *value*: computed value
            - *is_bin*, *is_oct*, *is_hex*, *is_dec*: flags set according to found type
            - *is_signed*: True when integer is signed
            - *size*: the size part

            Note: if there is no type information, type is _i32.
        """
        m = cls.TypedInteger.match(string)
        if not m:
            return None
        is_bin = m["value"].find("0b") != -1
        is_oct = m["value"].find("0o") != -1
        is_hex = m["value"].find("0x") != -1
        is_dec = not (is_bin or is_oct or is_hex)
        if m["type"]:
            assert not (minus and m["type"] == "_ui")
            is_signed = m["type"] == "_i" or minus
            size = int(m["size"])
        else:
            is_signed = True
            size = 32

        value = m["value"]
        if is_dec:
            value = int(value)
        elif is_bin:
            value = int(value[2:], 2)
        elif is_oct:
            value = int(value[2:], 8)
        elif is_hex:
            value = int(value[2:], 16)
        if minus:
            value = -value
        return IntegerTuple(value, is_bin, is_oct, is_hex, is_dec, is_signed, size)

    @classmethod
    def is_integer(cls, string: str) -> bool:
        """Check whether a string is a Swan integer.

        Parameters
        ----------
        string : str
            Integer value, as decimal, bin, octal, or hexadecimal, with
            or without type information.

        Returns
        -------
        bool
            True when string is an integer.
        """
        return cls.parse_integer(string) is not None

    @classmethod
    def parse_float(cls, string: str, minus: bool = False) -> Union[FloatTuple, None]:
        """Match a string representing a float and return
        a description of that float as a FloatTuple.

        Parameters
        ----------
        string : str
            String representing a float, with or without type information.
        minus : bool
            True when the value is preceded with a '-' minus operator.

        Returns
        -------
        FloatTuple or None
            If the string value matches SwanRE.TypedFloat pattern, a
            FloatTuple is returned. It is a namedtuple with fields:

            - *value*: computed value
            - *mantissa*: the mantissa part
            - *exp*: the exponent part
            - *size*: the size part

            Note: if there is no type information, type is _f32.
        """
        m = cls.TypedFloat.match(string)
        if not m:
            return None

        mantissa = m["mantissa"]
        exp = int(m["exp"]) if m["exp"] else 1
        size = int(m["size"]) if m["size"] else 32
        # We should use numpy for float32 / float64
        value = -float(m["value"]) if minus else float(m["value"])
        return FloatTuple(value, mantissa, exp, size)

    @classmethod
    def is_float(cls, string: str) -> bool:
        """Check whether a string is a Swan integer.

        Parameters
        ----------
        string : str
            Integer value, as decimal, bin, octal, or hexadecimal, with
            or without type information.

        Returns
        -------
        bool
            True when string is an integer.
        """
        return cls.parse_float(string) is not None

    @classmethod
    def is_numeric(cls, string: str) -> bool:
        """Check whether a string is a Swan numeric value,
        that is, an integer of float value.

        Parameters
        ----------
        string : str
            String to check.

        Returns
        -------
        bool
            True if string is a Swan numeric value.
        """
        return cls.is_integer(string) or cls.is_float(string)

    CharRe = re.compile(r"(?:'[--~]'|\\x[0-9a-fA-F]{'1,2})$")

    @classmethod
    def is_char(cls, string: str) -> bool:
        """Check whether a string is a Swan char value.

        Parameters
        ----------
        string : str
            String to check.

        Returns
        -------
        bool
            True if string is a Swan char value.
        """
        return cls.CharRe.match(string) is not None

    BoolRe = re.compile("(?:true|false)$")

    @classmethod
    def is_bool(cls, string: str) -> bool:
        """Check whether a string is a Swan boolean value.

        Parameters
        ----------
        string : str
            String to check.

        Returns
        -------
        bool
            True if string is a Swan char value
        """
        return cls.BoolRe.match(string) is not None


class Markup:  # numpydoc ignore=PR01
    """Class defining the markups used by the Swan serialization."""

    def __init__(self) -> None:
        # avoid the generic version with args and kwargs (introspection)
        pass

    NoMarkup = ""
    #: General syntax error.
    Syntax = "syntax"
    #: Incorrect variable declaration.
    Var = "var"
    #: Incorrect group declaration.
    Group = "group"
    #: Incorrect sensor declaration.
    Sensor = "sensor"
    #: Incorrect const declaration.
    Const = "const"
    #: Incorrect type declaration.
    Type = "type"
    #: Incorrect use declaration.
    Use = "use"
    #: Incorrect operator signature in interface.
    Signature = "signature"
    #: Textual operator or generic operator content. The content is re-parsed by the API.
    Text = "text"
    #: Textual operator with syntax error.
    SyntaxText = "syntax_text"
    #: Empty instance block body. This is an invalid construct, but it is needed for the editor.
    Empty = "empty"
    #: Protected instance id.
    Luid = "luid"
    #: Operator expression. Specific markup for the editor. The content is re-parsed by the API.
    OpExpr = "op_expr"
    #: Incorrect forward dimension.
    Dim = "dim"

    @staticmethod
    def to_str(text: str, is_protected: bool = True, markup: str = None) -> str:
        """Return *text* as the protected string {markup%text%markup} if required.

        Parameters
        ----------
        text : str
            Text to protect
        is_protected : bool, optional
            True when text shall be protected, by default True
        markup : str, optional
            Markup to use, by default None, resulting in using Markup.Syntax

        Returns
        -------
        str
            Protected string *{markup%text%markup}* if is_protected is True, else *text*.
        """
        if not is_protected:
            return text
        if markup is None:
            markup = Markup.Syntax
        return f"{{{markup}%{text}%{markup}}}"


class NumericKind(Enum):  # numpydoc ignore=PR01
    """Numeric kinds for generic type constraints."""

    # pylint: disable=invalid-name
    #: *numeric*
    Numeric = auto()
    #: *integer*
    Integer = auto()
    #: *signed*
    Signed = auto()
    #: *unsigned*
    Unsigned = auto()
    #: *float*
    Float = auto()

    @staticmethod
    def to_str(value: "NumericKind") -> str:
        return value.name.lower()


# ========================================================
# Common Constructs, used by other constructs


class Identifier(SwanItem, PragmaBase):  # numpydoc ignore=PR01
    """Class for identifier.

    An Identifier can be invalid if it was protected while saving it
    for some reason. In that case, the property *_is_valid_* is set to True.

    The class stores the pragmas associated with the Identifier.
    """

    IdentifierRE = re.compile(r"^[a-zA-Z]\w*$", re.ASCII)

    def __init__(
        self,
        value: str,
        pragmas: Optional[List[Pragma]] = None,
        comment: Optional[str] = None,
        is_name: Optional[bool] = False,
    ) -> None:
        """Identifier

        Parameters
        ----------
        value : str
            Identifier string.
        pragmas : List[Pragma]
            List of pragmas.
        comment : str, optional
            Comment, can be multiline, by default None.
        is_name : bool, optional
            True when Identifier is a name, aka. 'Identifier.
        """
        SwanItem.__init__(self)
        PragmaBase.__init__(self, pragmas)
        self._value = value
        self._comment = comment
        self._is_valid = Identifier.IdentifierRE.match(value) is not None
        self._is_name = is_name

    @property
    def is_valid(self) -> bool:
        """Return true when Identifier is valid."""
        return self._is_valid

    @property
    def must_be_protected(self) -> bool:
        """Returns true when Identifier must be protected (not valid)."""
        return not self.is_valid

    @property
    def value(self) -> str:
        """Identifier as a string."""
        return f"'{self._value}" if self._is_name else self._value

    @property
    def is_name(self) -> bool:
        """Return true when Identifier is a name."""
        return self._is_name

    @property
    def comment(self) -> str:
        """Comment part string."""
        return self._comment

    def __str__(self):
        """Generate a string representation of the Identifier. Pragmas are not included."""
        value = self.value if self.is_valid else Markup.to_str(self.value)
        return value


class PathIdentifier(SwanItem):  # numpydoc ignore=PR01
    """Class for path identifiers, i.e: P1::Id.

    The class manipulates the PathIdentifier as separate items.

    If the original path was protected (given as a string), the property *is_valid*
    is False, and the path is considered to be a single string and *is_protected* is True.

    path_id argument is:

    - a list of identifiers, for a valid path.
    - a string if the path was protected.
    """

    def __init__(self, path_id: Union[List[Identifier], str]) -> None:
        super().__init__()
        self._path_id = path_id
        self._is_valid = isinstance(path_id, list) and all(id.is_valid for id in path_id)

    def __eq__(self, other) -> bool:
        # Compare the two objects
        if not isinstance(other, PathIdentifier):
            return False

        # Compare path_id attributes, maybe it is a List[Identifier] or a string
        if isinstance(self.path_id, list) and isinstance(other.path_id, list):
            return self.as_string == other.as_string
        return self.path_id == other.path_id

    def __hash__(self):
        # Compute hash based on path_id
        if isinstance(self.path_id, list):
            # Convert list to tuple for hashing
            return hash(tuple(self.path_id))
        return hash(self.path_id)

    @staticmethod
    def from_string(path: str) -> "PathIdentifier":
        """Create a PathIdentifier from a string.

        Parameters
        ----------
        path : str
            String containing the path identifier.

        Returns
        -------
        PathIdentifier
            PathIdentifier instance.
        """
        return PathIdentifier([Identifier(p) for p in path.split("::")])

    @staticmethod
    def split(path: str) -> Tuple[str, str]:
        """Split a path into **path**::**name**.

        Parameters
        ----------
        path : str
            Path string.

        Returns
        -------
        Tuple[str, str]
            **path**, **name**.
        """
        parts = path.replace(" ", "").split("::")
        if len(parts) == 1:
            return "", parts[0]
        return "::".join(parts[:-1]), parts[-1]

    # id { :: id} * regexp, with spaces included
    PathIdentifierRe = re.compile(
        r"""^[a-zA-Z]\w*   # identifier
        (?:\s*::\s*[a-zA-Z]\w*)*  # sequence of ('-' identifier)
        $""",
        re.ASCII | re.VERBOSE,
    )

    @classmethod
    def is_valid_path(cls, path: str) -> bool:
        """Check if *path* is a valid path identifier, i.e.
        *id {:: id}*, with possible spaces around '::'.

        Parameters
        ----------
        path : str
            String containing the path identifier.

        Returns
        -------
        bool
            True when path is valid.
        """
        return cls.PathIdentifierRe.match(path)

    # id { - id} * regexp, with no spaces
    FilePathIdentifierRE = re.compile(
        r"""^[a-zA-Z]\w*   # identifier
                                    (?:-[a-zA-Z]\w*)*  # sequence of ('-' identifier)
                                    $""",
        re.ASCII | re.VERBOSE,
    )

    @classmethod
    def is_valid_file_path(cls, path: str) -> bool:
        """Check if *path* is a valid file path identifier, i.e.
        *id {- id}*, with no possible spaces around '-'.

        The path string is the basename of a module or an instance file.

        Parameters
        ----------
        path : str
            String containing the path identifier.

        Returns
        -------
        bool
            True when path is valid.
        """
        return cls.FilePathIdentifierRE.match(path)

    @property
    def is_valid(self) -> bool:
        """True when path is a sequence of Identifier."""
        return self._is_valid

    @property
    def is_protected(self) -> bool:
        """True when path is from a protected source, i.e., a string."""
        return not self.is_valid

    @property
    def path_id(self) -> Union[List[Identifier], str]:
        """PathId as a list of Identifier, or a string if protected."""
        return self._path_id

    @property
    def as_string(self) -> str:
        """Compute name by joining name parts with '::'."""
        return self.path_id if self.is_protected else "::".join([p.value for p in self.path_id])

    @property
    def pragmas(self) -> List[Pragma]:
        """List of Pragma associated with the path_id."""
        return [] if self.is_protected else self._path_id[0].pragmas

    def __str__(self) -> str:
        """Generate string representation of the PathIdentifier. Pragmas are not included."""
        return self.as_string


class ModuleItem(SwanItem):  # numpydoc ignore=PR01
    """Base class for module body item or module interface item."""

    def __init__(self) -> None:
        super().__init__()


class Declaration(SwanItem):  # numpydoc ignore=PR01
    """Base class for declarations."""

    def __init__(self, id: Identifier) -> None:
        super().__init__()
        self._id = id

    @property
    def id(self) -> Identifier:
        """Language item identifier."""
        return self._id

    def get_full_path(self) -> str:
        """Full path of Swan construct."""
        if self.owner is None:
            raise ScadeOneException("No owner")
        path = self.owner.get_full_path()
        id_str = self.id.value
        return f"{path}::{id_str}"


class Expression(SwanItem):  # numpydoc ignore=PR01
    """Base class for expressions."""

    def __init__(self) -> None:
        super().__init__()


class TypeExpression(SwanItem):  # numpydoc ignore=PR01
    """Base class for type expressions."""

    def __init__(self) -> None:
        super().__init__()

    @property
    def is_defined(self) -> bool:
        """True if type expression is a predefined type."""
        return False


class GroupTypeExpression(SwanItem):  # numpydoc ignore=PR01
    """Base class for group type expressions."""

    def __init__(self) -> None:
        super().__init__()


class Luid(SwanItem):  # numpydoc ignore=PR01
    """Class for LUID support.
    The '$' prefix is not saved (potentially removed at creation)."""

    LuidRE = re.compile(r"^\$?[a-zA-Z]\w*$", re.ASCII)

    def __init__(self, value: str) -> None:
        super().__init__()
        self._luid = value[1:] if value[0] == "$" else value

    @property
    def value(self) -> str:
        """Luid value."""
        return self._luid

    def as_luid(self) -> str:
        """Luid value with '$ prefix. Equivalent to str(luid)."""
        return f"${self.value}"

    @staticmethod
    def is_valid(luid: str) -> bool:  # numpydoc ignore=PR01
        """True when a luid is a valid LUID."""
        return Luid.LuidRE.match(luid) is not None

    def __str__(self) -> str:
        return self.as_luid()


class Lunum(SwanItem):  # numpydoc ignore=PR01
    """Class for LUNUM support: '#' is part of the LUNUM."""

    LunumRE = re.compile(r"^\#\d+$", re.ASCII)

    def __init__(self, value: str) -> None:
        super().__init__()
        self._luid = value

    @property
    def value(self) -> str:
        """Luid value as a string."""
        return self._luid

    @staticmethod
    def is_valid(lunum: str) -> bool:  # numpydoc ignore=PR01
        """True when a lunum is a valid LUNUM."""
        return Lunum.LunumRE.match(lunum) is not None

    def __str__(self) -> str:
        return self.value


class Variable(SwanItem):  # numpydoc ignore=PR01
    """Base class for Variable and ProtectedVariable."""

    def __init__(self) -> None:
        super().__init__()


class Equation(SwanItem):  # numpydoc ignore=PR01
    """Base class for equations."""

    def __init__(self) -> None:
        super().__init__()


# =============================================
# Protected Items
# =============================================


class ProtectedItem(SwanItem):  # numpydoc ignore=PR01
    """Base class for protected data. A protected data
    is a piece of Swan code enclosed between markups, mostly to store
    syntactically incorrect code. A protected data is enclosed within the pair
    ``{markup%`` .. ``%markup}``, where *markup* is defined by the
    regular expression: ['a'-'z' 'A'-'Z' 0-9 _]*.

    See :py:class:`Markup` for existing markups.
    """

    def __init__(self, data: str, markup: Optional[str] = Markup.Syntax) -> None:
        super().__init__()
        self._markup = markup
        self._data = data

    @property
    def is_protected(self) -> bool:
        """Tell if item is syntactically protected and provided as a string."""
        return True

    def has_markup(self, markup: str) -> bool:
        """Check if protected data has the specified *markup*.

        Parameters
        ----------
        markup : str
            String markup.

        Returns
        -------
        result: bool
            True when instance markup is same as parameter.
        """
        return self._markup == markup

    @property
    def data(self) -> str:
        """Protected data between markups.

        Returns
        -------
        str
            Protected data.
        """
        return self._data

    @property
    def markup(self) -> str:
        """Protection markup.

        Returns
        -------
        str
            Markup string.
        """
        return self._markup

    def __str__(self) -> str:
        """Generate a string representation of the protected data."""
        return Markup.to_str(self.data, markup=self.markup)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# cSpell: ignore vsize ndarray

import abc
from typing import Tuple, Optional, Any, List

from ansys.scadeone.core.common.exception import ScadeOneException
import ansys.scadeone.core.svc.simdata.core as core
import ansys.scadeone.core.svc.simdata.core.dll_wrap as dll_wrap
import ansys.scadeone.core.svc.simdata.defs as defs


class _SdFactory:
    @classmethod
    def load_type(cls, type_id: int) -> Optional[defs.Type]:
        # Predefined type => return it
        predefined_type = defs.PredefinedType.get(type_id)
        if predefined_type is not None:
            return predefined_type
        name = dll_wrap.sdt_get_name(type_id)
        type_class = dll_wrap.sdt_get_class(type_id)
        if type_class == core.TypeClass.STRUCT:
            nb_fields = dll_wrap.sdt_struct_get_n_fields(type_id)
            fields = []
            for i in range(0, nb_fields):
                field_name = dll_wrap.sdt_struct_get_field_name(type_id, i)
                field_offset = dll_wrap.sdt_struct_get_field_offset(type_id, i)
                field_type_id = dll_wrap.sdt_struct_get_field_type(type_id, i)
                field_type = cls.load_type(field_type_id)
                fields.append(defs.StructTypeField(field_name, field_offset, field_type))
            return defs.StructType(type_id, fields, name)
        elif type_class == core.TypeClass.ARRAY:
            base_type_id = dll_wrap.sdt_array_get_base_type(type_id)
            base_type = cls.load_type(base_type_id)
            dims = dll_wrap.sdt_array_get_dims(type_id)
            return defs.ArrayType(type_id, base_type, dims)
        elif type_class == core.TypeClass.ENUM:
            base_type_id = dll_wrap.sdt_enum_get_base_type(type_id)
            nb_values = dll_wrap.sdt_enum_get_n_values(type_id)
            enum_values = []
            for i in range(0, nb_values):
                tag_name = dll_wrap.sdt_enum_get_value_name(type_id, i)
                int_value = dll_wrap.sdt_enum_get_value_value(type_id, i)
                enum_values.append(defs.EnumTypeValue(tag_name, int_value))
            return defs.EnumType(type_id, defs.PredefinedType.get(base_type_id), enum_values, name)
        elif type_class == core.TypeClass.VARIANT:
            nb_constructors = dll_wrap.sdt_variant_get_n_constructors(type_id)
            constructors = []
            for i in range(0, nb_constructors):
                constructor_name = dll_wrap.sdt_variant_get_constructor_name(type_id, i)
                constructor_value_type_id = dll_wrap.sdt_variant_get_constructor_value_type(
                    type_id, i
                )
                constructor_value_type = (
                    None
                    if constructor_value_type_id == core.SDT_NONE
                    else cls.load_type(constructor_value_type_id)
                )
                constructors.append(
                    defs.VariantTypeConstructor(constructor_name, constructor_value_type)
                )
            return defs.VariantType(type_id, constructors, name)
        elif type_class == core.TypeClass.IMPORTED:
            mem_size = dll_wrap.sdt_get_size(type_id)
            vsize = dll_wrap.sdt_imported_is_variable_size(type_id)
            return defs.ImportedType(type_id, mem_size, vsize, name)
        return None

    @classmethod
    def load_element(
        cls, file: defs.FileBase, parent: Optional[defs.ElementBase], elem_id: int
    ) -> defs.ElementBase:
        elem_name = dll_wrap.sde_get_name(elem_id)
        elem_type_id = dll_wrap.sde_get_type(elem_id)
        elem_type = cls.load_type(elem_type_id)
        elem_kind = defs.ElementKind(dll_wrap.sde_get_kind(elem_id))
        elem = Element(file, elem_id, parent, elem_name, elem_type, elem_kind)
        return elem

    @classmethod
    def build_core_sd_value(cls, py_value: Any, sd_type: defs.Type) -> Optional[core.sd_value_t]:
        if py_value is None:
            return dll_wrap.sdd_value_create_none()
        if isinstance(sd_type, defs.PredefinedType):
            if sd_type.kind == defs.PredefinedTypeKind.CHAR:
                return dll_wrap.sdd_value_create_predef_char(py_value)
            elif sd_type.kind == defs.PredefinedTypeKind.BOOL:
                return dll_wrap.sdd_value_create_predef_bool(py_value)
            elif sd_type.kind == defs.PredefinedTypeKind.INT8:
                return dll_wrap.sdd_value_create_predef_int8(int(py_value))
            elif sd_type.kind == defs.PredefinedTypeKind.INT16:
                return dll_wrap.sdd_value_create_predef_int16(int(py_value))
            elif sd_type.kind == defs.PredefinedTypeKind.INT32:
                return dll_wrap.sdd_value_create_predef_int32(int(py_value))
            elif sd_type.kind == defs.PredefinedTypeKind.INT64:
                return dll_wrap.sdd_value_create_predef_int64(int(py_value))
            elif sd_type.kind == defs.PredefinedTypeKind.UINT8:
                return dll_wrap.sdd_value_create_predef_uint8(int(py_value))
            elif sd_type.kind == defs.PredefinedTypeKind.UINT16:
                return dll_wrap.sdd_value_create_predef_uint16(int(py_value))
            elif sd_type.kind == defs.PredefinedTypeKind.UINT32:
                return dll_wrap.sdd_value_create_predef_uint32(int(py_value))
            elif sd_type.kind == defs.PredefinedTypeKind.UINT64:
                return dll_wrap.sdd_value_create_predef_uint64(int(py_value))
            elif sd_type.kind == defs.PredefinedTypeKind.FLOAT32:
                return dll_wrap.sdd_value_create_predef_float32(float(py_value))
            elif sd_type.kind == defs.PredefinedTypeKind.FLOAT64:
                return dll_wrap.sdd_value_create_predef_float64(float(py_value))
            else:
                return None
        elif isinstance(sd_type, defs.StructType):
            if not isinstance(py_value, (list, tuple)) or len(py_value) != len(sd_type.fields):
                return None
            csd_values = []
            for py_value_item, field in zip(py_value, sd_type.fields):
                csd_value = cls.build_core_sd_value(py_value_item, field.sd_type)
                if csd_value is None:
                    return None
                csd_values.append(csd_value)
            return dll_wrap.sdd_value_create_list(len(csd_values), csd_values)
        elif isinstance(sd_type, defs.ArrayType):
            # for numpy ndarray:
            if hasattr(py_value, "shape"):
                if list(py_value.shape) != sd_type.dims:
                    return None
                array_cell_type = (
                    sd_type.base_type
                    if len(sd_type.dims) == 1
                    else defs.ArrayType(0, sd_type.base_type, sd_type.dims[1:])
                )
                csd_values = []
                for py_value_item, i in zip(py_value, range(0, sd_type.dims[0])):
                    csd_value = cls.build_core_sd_value(py_value_item, array_cell_type)
                    if csd_value is None:
                        return None
                    csd_values.append(csd_value)
                return dll_wrap.sdd_value_create_list(len(csd_values), csd_values)
            elif isinstance(py_value, (list, tuple)):
                if len(py_value) != sd_type.dims[0]:
                    return None
                array_cell_type = (
                    sd_type.base_type
                    if len(sd_type.dims) == 1
                    else defs.ArrayType(0, sd_type.base_type, sd_type.dims[1:])
                )
                csd_values = []
                for py_value_item, i in zip(py_value, range(0, sd_type.dims[0])):
                    csd_value = cls.build_core_sd_value(py_value_item, array_cell_type)
                    if csd_value is None:
                        return None
                    csd_values.append(csd_value)
                return dll_wrap.sdd_value_create_list(len(csd_values), csd_values)
            else:
                return None
        elif isinstance(sd_type, defs.EnumType):
            return dll_wrap.sdd_value_create_enum(py_value)
        elif isinstance(sd_type, defs.VariantType):
            csd_value = None
            if isinstance(py_value, tuple):
                if len(py_value) < 1 or len(py_value) > 2:
                    return None
                ctor_name = py_value[0]
                if len(py_value) == 2:
                    ctor = sd_type.find_constructor(ctor_name)
                    if ctor is None or ctor.value_type is None:
                        return None
                    csd_value = cls.build_core_sd_value(py_value[1], ctor.value_type)
            elif isinstance(py_value, str):
                ctor_name = py_value
            else:
                return None
            return dll_wrap.sdd_value_create_variant(ctor_name, csd_value)
        elif isinstance(sd_type, defs.ImportedType):
            return dll_wrap.sdd_value_create_imported(py_value)
        return None

    @classmethod
    def load_sd_value(cls, csd_value: core.sd_value_t) -> defs.Value:
        value_class = dll_wrap.sdd_value_get_class(csd_value)
        if value_class is core.DataClass.NONE:
            return defs.NoneValue()
        if value_class is core.DataClass.PREDEF:
            predef_id = dll_wrap.sdd_value_get_predef_type_id(csd_value)
            if predef_id == core.PredefinedType.CHAR:
                cui8 = dll_wrap.sdd_value_get_predef_uint8(csd_value)
                if cui8 is None:
                    raise ScadeOneException("cannot read char value")
                return defs.PredefinedCharValue(cui8)
            if predef_id == core.PredefinedType.BOOL:
                cui8 = dll_wrap.sdd_value_get_predef_uint8(csd_value)
                if cui8 is None:
                    raise ScadeOneException("cannot read bool value")
                return defs.PredefinedBoolValue(cui8)
            if predef_id == core.PredefinedType.INT8:
                ci8 = dll_wrap.sdd_value_get_predef_int8(csd_value)
                if ci8 is None:
                    raise ScadeOneException("cannot read int8 value")
                return defs.PredefinedInt8Value(ci8)
            if predef_id == core.PredefinedType.INT16:
                ci16 = dll_wrap.sdd_value_get_predef_int16(csd_value)
                if ci16 is None:
                    raise ScadeOneException("cannot read int16 value")
                return defs.PredefinedInt16Value(ci16)
            if predef_id == core.PredefinedType.INT32:
                ci32 = dll_wrap.sdd_value_get_predef_int32(csd_value)
                if ci32 is None:
                    raise ScadeOneException("cannot read int32 value")
                return defs.PredefinedInt32Value(ci32)
            if predef_id == core.PredefinedType.INT64:
                ci64 = dll_wrap.sdd_value_get_predef_int64(csd_value)
                if ci64 is None:
                    raise ScadeOneException("cannot read int64 value")
                return defs.PredefinedInt64Value(ci64)
            if predef_id == core.PredefinedType.UINT8:
                cui8 = dll_wrap.sdd_value_get_predef_uint8(csd_value)
                if cui8 is None:
                    raise ScadeOneException("cannot read uint8 value")
                return defs.PredefinedUInt8Value(cui8)
            if predef_id == core.PredefinedType.UINT16:
                cui16 = dll_wrap.sdd_value_get_predef_uint16(csd_value)
                if cui16 is None:
                    raise ScadeOneException("cannot read uint16 value")
                return defs.PredefinedUInt16Value(cui16)
            if predef_id == core.PredefinedType.UINT32:
                cui32 = dll_wrap.sdd_value_get_predef_uint32(csd_value)
                if cui32 is None:
                    raise ScadeOneException("cannot read uint32 value")
                return defs.PredefinedUInt32Value(cui32)
            if predef_id == core.PredefinedType.UINT64:
                cui64 = dll_wrap.sdd_value_get_predef_uint64(csd_value)
                if cui64 is None:
                    raise ScadeOneException("cannot read uint64 value")
                return defs.PredefinedUInt64Value(cui64)
            if predef_id == core.PredefinedType.FLOAT32:
                cf32 = dll_wrap.sdd_value_get_predef_float32(csd_value)
                if cf32 is None:
                    raise ScadeOneException("cannot read float32 value")
                return defs.PredefinedFloat32Value(cf32)
            if predef_id == core.PredefinedType.FLOAT64:
                cf64 = dll_wrap.sdd_value_get_predef_float64(csd_value)
                if cf64 is None:
                    raise ScadeOneException("cannot read float64 value")
                return defs.PredefinedFloat64Value(cf64)
        if value_class is core.DataClass.LIST:
            n_values = dll_wrap.sdd_value_get_list_n_values(csd_value)
            values = []
            for i in range(0, n_values):
                values.append(cls.load_sd_value(dll_wrap.sdd_value_get_list_value_at(csd_value, i)))
            return defs.ListValue(values)
        if value_class is core.DataClass.ENUM:
            return defs.EnumValue(dll_wrap.sdd_value_get_enum_name(csd_value))
        if value_class is core.DataClass.VARIANT:
            name = dll_wrap.sdd_value_get_variant_name(csd_value)
            csd_variant_value = dll_wrap.sdd_value_get_variant_value(csd_value)
            sd_value = None if csd_variant_value is None else cls.load_sd_value(csd_variant_value)
            return defs.VariantValue(name, sd_value)
        if value_class is core.DataClass.UNTYPED_VARIANT_CONSTRUCTOR:
            return defs.UntypedVariantConstructorValue()
        if value_class is core.DataClass.IMPORTED:
            byte_values = dll_wrap.sdd_value_get_imported_value(csd_value)
            return defs.ImportedValue(byte_values)
        return defs.NoneValue()


class Element(defs.ElementBase):
    """Class for simdata elements"""

    def __init__(
        self,
        file: defs.FileBase,
        elem_id: int,
        parent: Optional[defs.ElementBase],
        name: str,
        sd_type: Optional[defs.Type],
        kind: defs.ElementKind,
    ) -> None:
        super().__init__(file, parent, name, sd_type, kind)
        self._elem_id = elem_id
        # read children elements
        children_elements_ids = dll_wrap.sde_get_children(self._elem_id)
        if children_elements_ids is None:
            raise ScadeOneException('cannot read children elements for "{0}"'.format(self._name))
        for child_id in children_elements_ids:
            self._children_elements.append(_SdFactory.load_element(file, self, child_id))

    @property
    def elem_id(self) -> int:
        return self._elem_id

    @defs.ElementBase.name.setter
    def name(self, name: str) -> None:
        if dll_wrap.sde_set_name(self._elem_id, name) != core.SD_ERR_NONE:
            raise ScadeOneException(
                "cannot set element name for element with id {0}".format(self._elem_id)
            )
        self._name = name

    @defs.ElementBase.sd_type.setter
    def sd_type(self, sd_type: Optional[defs.Type]) -> None:
        if (
            dll_wrap.sde_set_type(self._elem_id, sd_type.type_id if sd_type else core.SDT_NONE)
            != core.SD_ERR_NONE
        ):
            raise ScadeOneException(
                "cannot set element type for element with id {0}".format(self._elem_id)
            )
        self._sd_type = sd_type

    @defs.ElementBase.kind.setter
    def kind(self, kind: defs.ElementKind) -> None:
        if dll_wrap.sde_set_kind(self._elem_id, core.SdeKind(kind)) != core.SD_ERR_NONE:
            raise ScadeOneException(
                "cannot set element kind for element with id {0}".format(self._elem_id)
            )
        self._kind = kind

    def add_child_element(
        self, name: str, sd_type: defs.Type = None, kind: defs.ElementKind = defs.ElementKind.NONE
    ) -> defs.ElementBase:
        """Add a child to element

        Parameters
        ----------
        name : str
            child element name
        sd_type : Type, optional
            child element type
        kind : ElementKind, optional
            child element kind

        Returns
        -------
        ElementBase
            Created child element
        """
        elem_id = dll_wrap.sde_create(
            self._elem_id, name, sd_type.type_id if sd_type else core.SDT_NONE, core.SdeKind(kind)
        )
        elem = Element(self._file, elem_id, self, name, sd_type, kind)
        self._children_elements.append(elem)
        return elem

    def remove_child_element(self, element: "Element") -> None:
        """Remove a child from element

        Parameters
        ----------
        element : Element
            element to remove

        Raises
        ------
        ScadeOneException
            child could not be removed
        """
        if dll_wrap.sde_remove(element.elem_id) != core.SD_ERR_NONE:
            raise ScadeOneException("cannot remove element with id {0}".format(element.elem_id))
        self._children_elements.remove(element)

    def append_value(self, py_value: Any) -> None:
        """Add a value to an element or appends to the last sequence of values if any.
        This method shall NOT be used to add a single value after a multiple cycles sequence
        as it would also be repeated alongside the previous sequence.
        Prefer using append_values([py_value]) instead in such case.

        Parameters
        ----------
        py_value : Any
            value that has a type corresponding to the element

        Raises
        ------
        ScadeOneException
            value is invalid or could not be added
        """
        csd_value = _SdFactory.build_core_sd_value(py_value, self.sd_type)
        if csd_value is None:
            raise ScadeOneException(f"invalid value: {py_value}")
        ret = dll_wrap.sdd_append_value(self._elem_id, csd_value)
        dll_wrap.sdd_value_close(csd_value)
        if ret != core.SD_ERR_NONE:
            raise ScadeOneException("cannot append value")

    def append_nones(self, count: int) -> None:
        """Add a sequence of none values to an element (undefined cycle values)

        Parameters
        ----------
        count : int
            number of Nones to add

        Raises
        ------
        ScadeOneException
            cannot create or append nones sequence
        """
        sequence = dll_wrap.sdd_sequence_create_nones(count)
        if not sequence:
            raise ScadeOneException("cannot create nones sequence")
        ret = dll_wrap.sdd_append_sequence(self.elem_id, sequence)
        dll_wrap.sdd_sequence_close(sequence)
        if ret != core.SD_ERR_NONE:
            raise ScadeOneException("cannot append nones sequence")

    def append_values(self, py_values: List[Any], repeat_factor: Optional[int] = 1) -> None:
        """Create a new sequence of one or multiple values.
        Do not use repeat factor for array and structure types.
        Do not use this for structure or array types that already have any value,
        use append_value() instead in such case.

        Parameters
        ----------
        py_values : List[Any]
            list of value(s) to add
        repeat_factor : Optional[int], optional
            number of times to add, by default 1 (no repeat), do not use with non scalar values

        Raises
        ------
        ScadeOneException
            Invalid repeat, none contained in values, invalid values, cannot create or append
            values sequence, used repeat factor for array or structure types
        """
        if repeat_factor < 1:
            raise ScadeOneException("invalid repeat factor")
        if isinstance(self.sd_type, defs.StructType) or isinstance(self.sd_type, defs.ArrayType):
            if repeat_factor > 1:
                raise ScadeOneException("no repeat factor allowed for structure and array types")
            if any(self.read_values()):
                raise ScadeOneException("structure/array type already has an existing sequence")
        csd_values = []
        for py_value in py_values:
            if py_value is None:
                raise ScadeOneException("values sequence cannot contain 'none'")
            csd_value = _SdFactory.build_core_sd_value(py_value, self.sd_type)
            if csd_value is None:
                raise ScadeOneException(f"invalid value: {py_value} from {py_values}")
            csd_values.append(csd_value)
        sequence = dll_wrap.sdd_sequence_create_values(
            self.sd_type.type_id, len(csd_values), csd_values, repeat_factor
        )
        if not sequence:
            raise ScadeOneException("cannot create values sequence")
        ret = dll_wrap.sdd_append_sequence(self._elem_id, sequence)
        dll_wrap.sdd_sequence_close(sequence)
        if ret != core.SD_ERR_NONE:
            raise ScadeOneException("cannot append values sequence")

    def read_values(
        self, start: Optional[int] = None, n: Optional[int] = None
    ) -> defs.Iterator[defs.Value]:
        """Read element values

        Parameters
        ----------
        start : int, optional
            start index for reading, beginning if not specified
        n : int, optional
            number of values to read, runs until end if not specified

        Yields
        ------
        Iterator[Value]
            each value read
        """
        value_iter = dll_wrap.sdd_value_iter_create(self._elem_id)
        if value_iter:
            if start:
                if dll_wrap.sdd_value_iter_seek(value_iter, start) != core.SD_ERR_NONE:
                    raise ScadeOneException(
                        "cannot seek index for element with id {1}: invalid start index {0}".format(
                            start, self._elem_id
                        )
                    )
            if n is None:
                n = core.SD_SIZE_NONE

            for i in range(0, n):
                csd_value = dll_wrap.sdd_value_iter_get_value(value_iter)
                if csd_value is None:
                    break
                yield _SdFactory.load_sd_value(csd_value)
                dll_wrap.sdd_value_close(csd_value)
            dll_wrap.sdd_value_iter_close(value_iter)

    def clear_values(self) -> None:
        """Clear all values of element

        Raises
        ------
        ScadeOneException
            values could not be cleared
        """
        value_iter = dll_wrap.sdd_value_iter_create(self._elem_id)
        if value_iter and dll_wrap.sdd_value_iter_clear_values(value_iter) != core.SD_ERR_NONE:
            raise ScadeOneException(
                "cannot clear values for element with id {0}".format(self._elem_id)
            )


class File(defs.FileBase):
    """Simdata files class"""

    __metaclass__ = abc.ABCMeta

    def __init__(self, file_id: int) -> None:
        super().__init__()
        self._file_id = file_id
        # read root elements
        elements_ids = dll_wrap.sde_get_children(self._file_id)
        if elements_ids is None:
            raise ScadeOneException("cannot read root elements")
        for elem_id in elements_ids:
            self._elements.append(_SdFactory.load_element(self, None, elem_id))

    def add_element(
        self, name: str, sd_type: defs.Type = None, kind: defs.ElementKind = defs.ElementKind.NONE
    ) -> defs.ElementBase:
        """Create and add an element to file

        Parameters
        ----------
        name : str
            element name
        sd_type : Type, optional
            type of element
        kind : ElementKind, optional
            kind of element

        Returns
        -------
        ElementBase
            The new element that was added to file
        """

        elem_id = dll_wrap.sde_create(
            self._file_id, name, sd_type.type_id if sd_type else core.SDT_NONE, core.SdeKind(kind)
        )
        elem = _SdFactory.load_element(self, None, elem_id)
        self._elements.append(elem)
        return elem

    def remove_element(self, element: Element) -> None:
        """Delete an element from file

        Parameters
        ----------
        element : Element
            element to delete

        Raises
        ------
        ScadeOneException
            element could not be found or deleted
        """
        if dll_wrap.sde_remove(element.elem_id) != core.SD_ERR_NONE:
            raise ScadeOneException("cannot remove element with id {0}".format(element.elem_id))
        self._elements.remove(element)

    def get_version(self) -> str:
        """Get the file version

        Returns
        -------
        str
            Version number
        """
        return dll_wrap.sdf_get_version(self._file_id)

    def close(self) -> None:
        """Close a file

        Raises
        ------
        ScadeOneException
            Could not close file or file was not opened
        """
        if dll_wrap.sdf_close(self._file_id) != core.SD_ERR_NONE:
            raise ScadeOneException("cannot close file with id {0}".format(self._file_id))

    def __str__(self) -> str:
        s = "*** Elements:\n"
        for elem in self._elements:
            s += str(elem)
        return s


def open_file(file_path: str) -> defs.FileBase:
    """Open a simdata file in read-only mode

    Parameters
    ----------
    file_path : str
        path of the file

    Returns
    -------
    FileBase
        The opened file

    Raises
    ------
    ScadeOneException
        file could not be opened
    """
    file_id = dll_wrap.sdf_open(file_path, core.FileOpenMode.READ)
    if file_id == core.SD_ID_INVALID:
        raise ScadeOneException('cannot open file "{0}"'.format(file_path))
    return File(file_id)


def create_file(file_path: str) -> defs.FileBase:
    """Create a simdata file

    Parameters
    ----------
    file_path : str
        path of the file

    Returns
    -------
    FileBase
        The created file

    Raises
    ------
    ScadeOneException
        file could not be created
    """
    file_id = dll_wrap.sdf_open(file_path, core.FileOpenMode.CREATE)
    if file_id == core.SD_ID_INVALID:
        raise ScadeOneException('cannot create file "{0}"'.format(file_path))
    return File(file_id)


def edit_file(file_path: str) -> defs.FileBase:
    """Open a simdata file in edit mode

    Parameters
    ----------
    file_path : str
        path of the file

    Returns
    -------
    FileBase
        The opened file

    Raises
    ------
    ScadeOneException
        The file could not be opened
    """
    file_id = dll_wrap.sdf_open(file_path, core.FileOpenMode.EDIT)
    if file_id == core.SD_ID_INVALID:
        raise ScadeOneException('cannot open file "{0}"'.format(file_path))
    return File(file_id)


def create_struct_type(fields: List[Tuple], name: str = "") -> defs.StructType:
    """Create a structure type

    Parameters
    ----------
    fields : List[Tuple]
        all fields of the structure
    name : str, optional
        structure name

    Returns
    -------
    StructType
        The created structure type

    Raises
    ------
    ScadeOneException
        structure type could not be created,
        arguments are invalid, or field names duplicate
    """
    if (
        not isinstance(fields, list)
        or len(fields) == 0
        or not all(
            isinstance(f, tuple)
            and len(f) == 2
            and isinstance(f[0], str)
            and isinstance(f[1], defs.Type)
            for f in fields
        )
    ):
        raise ScadeOneException("invalid arguments in creation of struct type")
    struct_fields = []
    field_names = []
    for f in fields:
        field_name = f[0]
        field_type = f[1]
        if not isinstance(field_name, str):
            raise ScadeOneException("argument type error in creation of struct type")
        if field_name in field_names:
            raise ScadeOneException(
                'cannot add again field "{0}" to struct type'.format(field_name)
            )
        field_names.append(field_name)
        struct_fields.append(defs.StructTypeField(field_name, core.SD_SIZE_NONE, field_type))

    struct_type_id = dll_wrap.sdt_struct_create(core.SD_SIZE_NONE)
    if struct_type_id == core.SD_ID_INVALID:
        raise ScadeOneException("cannot create structure type")
    dll_wrap.sdt_set_name(struct_type_id, name)
    for struct_field in struct_fields:
        err = dll_wrap.sdt_struct_add_field(
            struct_type_id, struct_field.name, struct_field.offset, struct_field.sd_type.type_id
        )
        if err != core.SD_ERR_NONE:
            raise ScadeOneException(
                'cannot create field "{0}" of structure type'.format(struct_field.name)
            )
    return defs.StructType(struct_type_id, struct_fields, name)


def create_array_type(base_type: defs.Type, dims: List[int], name: str = "") -> defs.ArrayType:
    """Create a multi dimensional array type

    Parameters
    ----------
    base_type : Type
        type of array
    dims : List[int]
        array dimensions
    name : str, optional
        array name

    Returns
    -------
    ArrayType
        Created array type

    Raises
    ------
    ScadeOneException
        array type could not be created, argument error or none type passed
    """
    if base_type is None:
        raise ScadeOneException("cannot create array type. Base type shall not be None")
    if (
        not isinstance(base_type, defs.Type)
        or not isinstance(dims, list)
        or not all(isinstance(x, int) for x in dims)
        or len(dims) == 0
    ):
        raise ScadeOneException("argument type error in creation of array type")
    array_type_id = dll_wrap.sdt_array_create(base_type.type_id, len(dims), dims)
    if array_type_id == core.SD_ID_INVALID:
        raise ScadeOneException("cannot create array type")
    dll_wrap.sdt_set_name(array_type_id, name)
    return defs.ArrayType(array_type_id, base_type, dims, name)


def create_enum_type(values: List[str], name: str = "") -> defs.EnumType:
    """Create an enumeration type

    Parameters
    ----------
    values : List[str]
        values of the enumeration type
    name : str, optional
        name of enumeration type

    Returns
    -------
    EnumType
        Created enumeration type

    Raises
    ------
    ScadeOneException
        could not create enumeration type, create a value of passed argument
        or invalid arguments in creation
    """
    if len(values) == 0 or not all(isinstance(arg, str) for arg in values):
        raise ScadeOneException("argument type error in creation of enum type")

    enum_type_id = dll_wrap.sdt_enum_create(defs.Int64.type_id)
    if enum_type_id == core.SD_ID_INVALID:
        raise ScadeOneException("cannot create enum type")
    dll_wrap.sdt_set_name(enum_type_id, name)
    int_value = 0
    enum_values = []
    for value_name in values:
        err = dll_wrap.sdt_enum_add_value(enum_type_id, value_name, int_value)
        if err != core.SD_ERR_NONE:
            raise ScadeOneException('cannot create value "{0}" of enum type'.format(value_name))
        int_value = int_value + 1
        enum_values.append(defs.EnumTypeValue(value_name, int_value))
    return defs.EnumType(enum_type_id, defs.Int64, enum_values, name)


def create_variant_type(constructors: List[Tuple], name: str = "") -> defs.VariantType:
    """Create a variant type

    Parameters
    ----------
    constructors : List[Tuple]
        constructors of the variant type
    name : str, optional
        name of variant type

    Returns
    -------
    VariantType
        Created variant type

    Raises
    ------
    ScadeOneException
        variant type could not be created, invalid arguments or duplicate constructor names
    """
    if (
        not isinstance(constructors, list)
        or len(constructors) == 0
        or not all(
            isinstance(c, tuple)
            and len(c) == 2
            and isinstance(c[0], str)
            and (c[1] is None or isinstance(c[1], defs.Type))
            for c in constructors
        )
    ):
        raise ScadeOneException("invalid arguments in creation of variant type")

    variant_constructors = []
    constructor_names = []
    for constructor in constructors:
        constructor_name = constructor[0]
        constructor_value_type = constructor[1]
        if (
            not isinstance(constructor_name, str)
            or not isinstance(constructor_value_type, defs.Type)
            and constructor_value_type
        ):
            raise ScadeOneException("argument type error in creation of variant type")
        if constructor_name in constructor_names:
            raise ScadeOneException(
                'cannot add again constructor "{0}" to variant type'.format(constructor_name)
            )
        constructor_names.append(constructor_name)
        variant_constructors.append(
            defs.VariantTypeConstructor(constructor_name, constructor_value_type)
        )

    variant_type_id = dll_wrap.sdt_variant_create(core.SD_SIZE_NONE)
    if variant_type_id == core.SD_ID_INVALID:
        raise ScadeOneException("cannot create variant type")
    dll_wrap.sdt_set_name(variant_type_id, name)
    for variant_constructor in variant_constructors:
        err = dll_wrap.sdt_variant_add_constructor(
            variant_type_id,
            variant_constructor.name,
            core.SD_SIZE_NONE,
            core.SD_TAG_VALUE_NONE,
            core.SD_SIZE_NONE,
            (
                variant_constructor.value_type.type_id
                if variant_constructor.value_type
                else core.SDT_NONE
            ),
        )
        if err != core.SD_ERR_NONE:
            raise ScadeOneException(
                'cannot create constructor "{0}" of variant type'.format(variant_constructor.name)
            )
    return defs.VariantType(variant_type_id, variant_constructors, name)


def create_imported_type(mem_size: int, name: str = "") -> defs.ImportedType:
    """Create an imported type

    Parameters
    ----------
    mem_size : int
        memory size
    name : str, optional
        imported type name

    Returns
    -------
    ImportedType
        Created imported type

    Raises
    ------
    ScadeOneException
        could not create imported type, argument type error or argument type error
    """
    if not isinstance(mem_size, int) or mem_size is None:
        raise ScadeOneException("argument type error in creation of imported type")
    imported_type_id = dll_wrap.sdt_imported_create(mem_size)
    if imported_type_id == core.SD_ID_INVALID:
        raise ScadeOneException("cannot create imported type")
    dll_wrap.sdt_set_name(imported_type_id, name)
    return defs.ImportedType(imported_type_id, mem_size, False, None, None, name)


def create_vsize_imported_type(
    mem_size: int,
    pfn_vsize_get_bytes_size: defs.PfnVsizeGetBytesSize,
    pfn_vsize_to_bytes: defs.PfnVsizeToBytes,
    name: str = "",
) -> defs.ImportedType:
    if not isinstance(mem_size, int):
        raise ScadeOneException("argument type error in creation of imported type")
    imported_type_id = dll_wrap.sdt_vsize_imported_create(
        mem_size, pfn_vsize_get_bytes_size, pfn_vsize_to_bytes
    )
    if imported_type_id == core.SD_ID_INVALID:
        raise ScadeOneException("cannot create imported type")
    dll_wrap.sdt_set_name(imported_type_id, name)
    return defs.ImportedType(
        imported_type_id, mem_size, True, pfn_vsize_get_bytes_size, pfn_vsize_to_bytes, name
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.dotnet import load_dll

if TYPE_CHECKING:
    import ANSYS.SONE.Build.Toolkit as Toolkit  # type:ignore


class BuildResult:
    """Build result.

    Parameters
    ----------
    net_result : Toolkit.BuildResult
        .NET build result.
    """

    def __init__(self, net_result: "Toolkit.BuildResult") -> None:
        self._success = net_result.Success
        self._messages = [str(m) for m in net_result.Messages]

    @property
    def success(self) -> bool:
        return self._success

    @property
    def messages(self) -> List[str]:
        return self._messages

    def __str__(self) -> str:
        res = "Build " + ("succeeded" if self._success else "failed")
        if self._messages:
            res += "\nBuild messages: "
            for m in self._messages:
                res += "\n - " + str(m)
        return res


class BuildConfig:
    """Build configuration for building C source files.

    Attributes
    ----------
    working_dir : str
        Working directory.
    c_files : List[str]
        List of C source files.
    h_files : List[str]
        List of header files.
    o_files : List[str]
        List of object files.
    include_dirs : List[str]
        List of include directories.
    proto_files : List[str]
        List of protocol buffer files.
    preproc_defs : List[str]
        List of preprocessor definitions.
    lib_files : List[str]
        List of library files.
    targets : List[Target]
        List of build targets.
    mingw_dir : Optional[str]
        Path to the MinGW directory.
    compiler_config : Optional[CompilerConfig]
        Compiler configuration.
    incremental : bool
        Incremental build.
    """

    def __init__(self) -> None:
        self.working_dir: str = ""
        self.c_files: List[str] = []
        self.h_files: List[str] = []
        self.o_files: List[str] = []
        self.include_dirs: List[str] = []
        self.proto_files: List[str] = []
        self.preproc_defs: List[str] = []
        self.lib_files: List[str] = []
        self.targets: List[Target] = []
        self.mingw_dir: Optional[str] = None
        self.compiler_config: Optional[CompilerConfig] = None
        self.incremental: bool = False


class BuildSystem:
    """Build an executable or a shared library from C source files.

    Parameters
    ----------
    sone_install_dir : Union[Path, str]
        Path to the Scade One installation directory.
    """

    def __init__(self, sone_install_dir: Union[Path, str]) -> None:
        self._result = None
        if not sone_install_dir:
            raise ScadeOneException("Scade One installation directory not provided.")
        if isinstance(sone_install_dir, str):
            self._sone_install_dir = Path(sone_install_dir)
        else:
            self._sone_install_dir = sone_install_dir
        if not self._sone_install_dir.exists():
            raise ScadeOneException(
                f"Scade One installation directory not found: {self._sone_install_dir}"
            )
        self._register_build_system_dll()

    @property
    def result(self) -> BuildResult:
        """Return the build result."""
        return self._result

    def build(self, config: BuildConfig) -> BuildResult:
        """Build the C source files according to the build configuration.

        Parameters
        ----------
        config : BuildConfig
            Build configuration.

        Returns
        -------
        BuildResult
            Build result.
        """
        import ANSYS.SONE.Build.Toolkit as Toolkit  # type:ignore

        net_config = NetBuildConfig(config, self._sone_install_dir)
        bld = Toolkit.Builder()
        self._result = BuildResult(bld.Build(net_config.config))
        return self._result

    def _register_build_system_dll(self) -> None:
        simulator_dir = self._sone_install_dir / "tools/simulator"
        if not simulator_dir.exists():
            raise ScadeOneException(f"Simulator directory not found: {simulator_dir}")
        references = ["ANSYS.SONE.Build.Toolkit", "System.Collections"]
        load_dll(simulator_dir, references)


class TargetKind(Enum):
    """Kind of build target."""

    #: Executable target.
    EXECUTABLE = auto()

    #: Shared library target.
    SHARED_LIBRARY = auto()


class Target:
    """Build target.

    Parameters
    ----------
    base_name : str
        Name of the target.
    kind : TargetKind
        Kind of the target.
    """

    def __init__(self, base_name: str, kind: TargetKind):
        self._base_name = base_name
        self._kind = kind

    @property
    def base_name(self) -> str:
        return self._base_name

    @property
    def kind(self) -> TargetKind:
        return self._kind


class CompilerConfig:
    """Compiler configuration.

    Parameters
    ----------
    compiler_mk_path : str
        Path to the compiler configuration file.
    make_cmd_path : str
        Path to the make command.
    """

    def __init__(self, compiler_mk_path: str, make_cmd_path: str) -> None:
        self._compiler_mk_path = compiler_mk_path
        self._make_cmd_path = make_cmd_path

    @property
    def compiler_mk_path(self) -> str:
        return self._compiler_mk_path

    @property
    def make_cmd_path(self) -> str:
        return self._make_cmd_path


class NetBuildConfig:
    """Convert a Python BuildConfig to a .NET BuildConfig.

    Parameters
    ----------
    config : BuildConfig
        Python build configuration.
    """

    def __init__(self, config: BuildConfig, sone_install_dir: Optional[Path] = None) -> None:
        import ANSYS.SONE.Build.Toolkit as Toolkit  # type:ignore
        from System.Collections.Generic import List as NetList  # type:ignore

        self._net_config = Toolkit.BuildConfig()
        self._net_config.WorkingDir = config.working_dir
        self._net_config.CFiles = self._to_net_str_list(config.c_files)
        self._net_config.HFiles = self._to_net_str_list(config.h_files)
        self._net_config.OFiles = self._to_net_str_list(config.o_files)
        self._net_config.IncludeDirs = self._to_net_str_list(config.include_dirs)
        self._net_config.ProtoFiles = self._to_net_str_list(config.proto_files)
        self._net_config.PreprocDefs = self._to_net_str_list(config.preproc_defs)
        self._net_config.LibFiles = self._to_net_str_list(config.lib_files)
        self._net_config.Targets = NetList[Toolkit.Target]()
        for target in config.targets:
            self._net_config.Targets.Add(NetTarget(target).target)
        if config.mingw_dir:
            self._net_config.MingwDir = str(config.mingw_dir)
        elif sone_install_dir:
            self._net_config.MingwDir = str(sone_install_dir / "contrib/mingw64")
        else:
            raise ScadeOneException("MinGW directory not provided.")
        if config.compiler_config:
            self._net_config.CompilerConfig = NetCompilerConfig(config.compiler_config).config
        self._net_config.Incremental = config.incremental

    @property
    def config(self) -> "Toolkit.BuildConfig":
        """Return the .NET build configuration."""
        return self._net_config

    @staticmethod
    def _to_net_str_list(
        py_str_list: List[str],
    ):  # TODO: fix adding typing return -> NetList[NetString]
        """Convert a Python list of strings to a .NET list of strings."""
        from System import String as NetString  # type:ignore
        from System.Collections.Generic import List as NetList  # type:ignore

        net_list = NetList[NetString]()
        for s in py_str_list:
            net_list.Add(s)
        return net_list


class NetTarget:
    """Convert a Python Target to a .NET Target.

    Parameters
    ----------
    target : Target
        Python target.
    """

    def __init__(self, target: Target) -> None:
        import ANSYS.SONE.Build.Toolkit as Toolkit  # type:ignore

        kind = None
        if target.kind == TargetKind.EXECUTABLE:
            kind = Toolkit.TargetKind.Executable
        elif target.kind == TargetKind.SHARED_LIBRARY:
            kind = Toolkit.TargetKind.SharedLibrary

        self._net_target = Toolkit.Target(target.base_name, kind)

    @property
    def target(self) -> "Toolkit.Target":
        """Return the .NET target."""
        return self._net_target


class NetCompilerConfig:
    """Convert a Python CompilerConfig to a .NET CompilerConfig.

    Parameters
    ----------
    config : CompilerConfig
        Python compiler configuration.
    """

    def __init__(self, config: CompilerConfig) -> None:
        import ANSYS.SONE.Build.Toolkit as Toolkit  # type:ignore

        self._net_compiler_config = Toolkit.CompilerConfig(
            config.compiler_mk_path, config.make_cmd_path
        )

    @property
    def config(self) -> "Toolkit.CompilerConfig":
        """Return the .NET compiler configuration."""
        return self._net_compiler_config



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the classes for forward expression

"""

from enum import Enum, auto
from typing import List, Optional, Union

from ansys.scadeone.core.common.exception import ScadeOneException
import ansys.scadeone.core.swan.common as common
import ansys.scadeone.core.swan.scopes as scopes


# Forward Expression
# ======================================================================
class ForwardLHS(common.SwanItem):  # numpydoc ignore=PR01
    """**forward** construct:

    *current_lhs* ::= *id* | [ *current_lhs* ]"""

    def __init__(self, lhs: Union[common.Identifier, "ForwardLHS"]) -> None:
        super().__init__()
        self._lhs = lhs

    @property
    def lhs(self) -> Union[common.Identifier, "ForwardLHS"]:
        """Returns current lhs as an Identifier or a ForwardLHS."""
        return self._lhs

    @property
    def is_id(self) -> bool:
        """True when current lhs is an ID."""
        return isinstance(self.lhs, common.Identifier)


class ForwardElement(common.SwanItem):  # numpydoc ignore=PR01
    """Forward current element:

    *current_elt* ::= *current_lhs* = *expr* ;"""

    def __init__(self, lhs: ForwardLHS, expr: common.Expression) -> None:
        super().__init__()
        self._lhs = lhs
        self._expr = expr

    @property
    def lhs(self) -> ForwardLHS:
        """Current element."""
        return self._lhs

    @property
    def expr(self) -> common.Expression:
        """Current element expression."""
        return self._expr


class ForwardDim(common.SwanItem):  # numpydoc ignore=PR01
    """**forward** construct dimension:

    *dim* ::= << *expr* >> [[ **with** (( << *id* >> | *current_elt* )) {{ *current_elt* }} ]]

    Note that:

    - there may be no **with** part,
    - or it is an ID followed by a possible empty list,
    - or if no ID, at least one *current_element*.

    The *is_valid()* method checks for that property.

    Parameters
    ----------
    expr: common.Expression
       Dimension expression.

    id: common.Identifier (optional)
       **with** ID.

    elems: List[ForwardElement] (optional)
       **with** elements part.

    protected: str (optional)
        Content of the dimension if it syntactically incorrect.
        In that case, all other parameters are None.

    """

    def __init__(
        self,
        expr: Optional[common.Expression] = None,
        dim_id: Optional[common.Identifier] = None,
        elems: Optional[List[ForwardElement]] = None,
        protected: Optional[str] = None,
    ) -> None:
        super().__init__()
        self._expr = expr
        self._dim_id = dim_id
        self._elems = elems
        self._is_protected = protected is not None
        self._protected = protected

    @property
    def is_protected(self) -> bool:
        """True when dimension is syntactically incorrect and protected."""
        return self._is_protected

    @property
    def value(self) -> str:
        """Protected dimension content."""
        return self._protected

    @property
    def expr(self) -> common.Expression:
        """**forward** dimension expression."""
        return self._expr

    @property
    def dim_id(self) -> Union[common.Identifier, None]:
        """**forward** dimension ID, or None."""
        return self._dim_id

    @property
    def elems(self) -> Union[List[ForwardElement], None]:
        """**forward** dimension elements or None."""
        return self._elems

    @property
    def is_valid(self) -> bool:
        """Returns True when ID is given, or list of elements is not empty."""
        if self.is_protected:
            return False
        if self.dim_id:
            return True
        if (self.elems is None) or (len(self.elems) > 0):
            return True
        return False

    @property
    def protected(self) -> str:
        """Returns protected form as a string if dimension is syntactically invalid."""
        return self._protected


class ForwardLastDefault(common.SwanItem):  # numpydoc ignore=PR01
    """**forward** construct: *last_default*.

    *last_default* ::= **last** = *expr*
                   | **default** = *expr*
                   | **last** = *expr* **default** = *expr*
                   | **last** = **default** = *expr*

    Parameters
    ----------
    last: common.Expression (optional)
        **last** expression.

    default: common.Expression (optional)
        **default** expression.

    shared: common.Expression (optional)
        **last** and **default** share the same expression.
        *shared* cannot be used with *last* or *default*.
    """

    def __init__(
        self,
        last: Optional[common.Expression] = None,
        default: Optional[common.Expression] = None,
        shared: Optional[common.Expression] = None,
    ) -> None:
        super().__init__()
        self._last = last
        self._default = default
        self._shared = shared
        if (shared and (last or default)) or not (shared or last or default):
            raise ScadeOneException("Invalid ForwardLastDefault construction")

    @property
    def is_shared(self) -> bool:
        """True when **last** = **default** = *expr*."""
        return self._shared is not None

    @property
    def last(self) -> Union[common.Expression, None]:
        """Returns **last** expression or shared one."""
        if self._last:
            return self._last
        return self._shared

    @property
    def default(self) -> Union[common.Expression, None]:
        """Returns **default** expression or shared one."""
        if self._default:
            return self._default
        return self._shared

    @property
    def shared(self) -> Union[common.Expression, None]:
        """Return **shared** expression."""
        return self._shared


class ForwardItemClause(common.SwanItem):  # numpydoc ignore=PR01
    """**forward** construct:

    *item_clause* ::= *id* [[ : *last_default* ]]"""

    def __init__(
        self, id: common.Identifier, last_default: Optional[ForwardLastDefault] = None
    ) -> None:
        super().__init__()
        self._id = id
        self._last_default = last_default

    @property
    def id(self) -> common.Identifier:
        """Item_clause identifier."""
        return self._id

    @property
    def last_default(self) -> Union[ForwardLastDefault, None]:
        """Item_clause last default."""
        return self._last_default


class ForwardArrayClause(common.SwanItem):  # numpydoc ignore=PR01
    """**forward** construct:

    *returns_clause* ::= (( *item_clause* | *array_clause* ))
    *array_clause* ::= [ *returns_clause* ]
    """

    def __init__(self, return_clause: Union[ForwardItemClause, "ForwardArrayClause"]) -> None:
        super().__init__()
        self._return_clause = return_clause

    @property
    def return_clause(self) -> Union[ForwardItemClause, "ForwardArrayClause"]:
        """Return *array_clause* content."""
        return self._return_clause


class ForwardReturnItem(common.SwanItem):  # numpydoc ignore=PR01
    """Base class for *returns_item*."""

    def __init__(self) -> None:
        super().__init__()


class ForwardReturnItemClause(ForwardReturnItem):  # numpydoc ignore=PR01
    """**forward** construct: *returns_item* ::= *item_clause*."""

    def __init__(self, item_clause: ForwardItemClause) -> None:
        super().__init__()
        self._item_clause = item_clause

    @property
    def item_clause(self) -> ForwardItemClause:
        """Item clause."""
        return self._item_clause


class ForwardReturnArrayClause(ForwardReturnItem):  # numpydoc ignore=PR01
    """**forward** construct:

    *returns_item* ::= [[ *id* = ]] *array_clause*"""

    def __init__(
        self,
        array_clause: ForwardArrayClause,
        return_id: Optional[common.Identifier] = None,
    ) -> None:
        super().__init__()
        self._array_clause = array_clause
        self._return_id = return_id

    @property
    def array_clause(self) -> ForwardArrayClause:
        """Array clause."""
        return self._array_clause

    @property
    def return_id(self) -> Union[common.Identifier, None]:
        """Identifier of clause, or None."""
        return self._return_id


class ProtectedForwardReturnItem(common.ProtectedItem, ForwardReturnItem):  # numpydoc ignore=PR01
    """**forward** construct: protected *returns_item* with {syntax% ... %syntax} markup."""

    def __init__(self, data: str) -> None:
        super().__init__(data)


class ForwardState(Enum):  # numpydoc ignore=PR01  # numpydoc ignore=PR01
    """Forward state enumeration."""

    # pylint: disable=invalid-name
    Nothing = auto()
    Restart = auto()
    Resume = auto()

    @staticmethod
    def to_str(value: "ForwardState") -> str:
        if value == ForwardState.Nothing:
            return ""
        return value.name.lower()


class ForwardBody(common.SwanItem):  # numpydoc ignore=PR01
    """
    **forward** construct:

    fwd_body ::= [[ unless expr ]] scope_sections [[ until expr ]]
    """

    def __init__(
        self,
        body: List[scopes.ScopeSection],
        unless_expr: Optional[common.Expression] = None,
        until_expr: Optional[common.Expression] = None,
    ) -> None:
        super().__init__()
        self._body = body
        self._unless_expr = unless_expr
        self._until_expr = until_expr
        common.SwanItem.set_owner(self, body)

    @property
    def body(self) -> List[scopes.ScopeSection]:
        return self._body

    @property
    def unless_expr(self) -> Optional[common.Expression]:
        return self._unless_expr

    @property
    def until_expr(self) -> Optional[common.Expression]:
        return self._until_expr


class Forward(common.Expression):  # numpydoc ignore=PR01
    """Forward expression:

    | *fwd_expr* ::= **forward** [[ *luid*]] [[ (( **restart** | **resume** )) ]] {{ *dim* }}+
    |                *fwd_body* **returns** ( *returns_group* )
    | *returns_group* ::= [[ *returns_item* {{ , *returns_item* }} ]]
    """

    def __init__(
        self,
        state: ForwardState,
        dimensions: List[ForwardDim],
        body: ForwardBody,
        returns: List[ForwardReturnItem],
        luid: Optional[common.Luid] = None,
    ) -> None:
        super().__init__()
        self._state = state
        self._dimensions = dimensions
        self._body = body
        self._returns = returns
        self._luid = luid
        common.SwanItem.set_owner(self, body)

    @property
    def state(self) -> ForwardState:
        return self._state

    @property
    def dimensions(self) -> List[ForwardDim]:
        return self._dimensions

    @property
    def body(self) -> ForwardBody:
        return self._body

    @property
    def returns(self) -> List[ForwardReturnItem]:
        return self._returns

    @property
    def luid(self) -> Union[common.Luid, None]:
        return self._luid



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the :py:class:`Parser` which is a proxy to the Scade One F# parser.
The _Parser_ class offers methods to parse a complete Swan code (interface or module),
or a declaration, or an expression.

It relies on the `ansys.scadeone.core.model.dotnet` and `ansys.scadeone.model.pyofast` modules
to interface with the dotnet DLLs and to transform F# data structure into the
`ansys.scadeone.swan` python classes.
"""

import logging
from typing import Union

# dotnet configuration
import ansys.scadeone.core.model.dotnet  # noqa

# pylint: disable-next=import-error
from ANSYS.SONE.Core.Toolkit.Logging import ILogger  # type:ignore

# pylint: disable-next=import-error
from ANSYS.SONE.Infrastructure.Services.Serialization.BNF.Parsing import (  # type:ignore
    ParserTools,  # noqa
    Reader,  # noqa
)

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.storage import SwanStorage

import ansys.scadeone.core.swan as S

from .parser import Parser
from .pyofast import (
    declarationOfAst,
    equationOfAst,
    expressionOfAst,
    interfaceOfAst,
    moduleOfAst,
    operatorBlockOfAst,
    operatorExprOfAst,
    operatorOfAst,
    signatureOfAst,
    scopeSectionOfAst,
)

# version as a comment string for swan files
SwanVersion = ParserTools.SwanVersion
# version as a comment string for swant files
SwanTestVersion = ParserTools.SwanTestVersion
# dictionary of Swan versions
VersionMap = ParserTools.VersionInfos


class ParserLogger(ILogger):
    """Logger class for the parser. An instance of the
    class is given to the F# parser to get the logging information
    in Python world.

    The class only implements the methods from ILogger that may be called
    from the parser.

    Parameters
    ----------
    ILogger : ILogger
        C# interface
    """

    def __init__(self, logger: logging.Logger) -> None:
        self._logger = logger

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    # https://stackoverflow.com/questions/49736531/implement-a-c-sharp-interface-in-python-for-net
    __namespace__ = "MyPythonLogger"

    def _log(self, category, message, log_fn):
        log_fn(f"{category}: {message}")

    # pylint: disable=invalid-name
    def Info(self, category, message):
        self._log(category, message, self.logger.info)

    def Warning(self, category, message):
        self._log(category, message, self.logger.warning)

    def Error(self, category, message):
        self._log(category, message, self.logger.error)

    def Exception(self, category, message):
        self._log(category, message, self.logger.exception)

    def Debug(self, category, message):
        self._log(category, message, self.logger.debug)


class SwanParser(Parser):
    """The parser class is a proxy to the F# methods implemented
    by the parser.
    """

    def __init__(self, logger: logging.Logger) -> None:
        self._logger = ParserLogger(logger)

    def _parse(self, rule_fn, swan: SwanStorage, parse_error_ok: bool = False):
        """Call F# parser with a given rule

        Parameters
        ----------
        rule_fn : Callable
            Parser rule function
        swan : SwanStorage
            Swan code to parse

        Returns
        -------
        tuple
            Parser value, or None if parse_error_ok is True and parse error occurs.

        Raises
        ------
        ScadeOneException
            Raised when an error occurs during parsing, unless *parse_error_ok* is True.
            Raised when an internal error occurs.
        """
        # save current parser for pyofast methods
        Parser.set_current_parser(self)
        Parser.set_source(swan)

        try:
            result = rule_fn(swan.source, swan.content(), self._logger)
        except Reader.ParseError as e:
            if parse_error_ok:
                return None
            raise ScadeOneException(f"Parser: {e.Message}")
        except Exception as e:
            raise ScadeOneException(f"Internal: {e}")
        return result

    def module_body(self, source: SwanStorage) -> S.ModuleBody:
        """Parse a Swan module from a SwanStorage object.

            The *content()* method is called to get the code.

            The *name* property is used to set the module identifier.

        Parameters
        ----------
        source : SwanStorage
            Swan module (.swan)

        Returns
        -------
        ModuleBody
            Instance of ModuleBody.
        """
        source.check_swan_version()
        result = self._parse(Reader.parse_body, source)
        return moduleOfAst(source.name, result.Item1)

    def test_harness(self, source: SwanStorage) -> S.TestModule:
        """Parse a test harness from a SwanStorage object.

            The *content()* method is called to get the code.

            The *name* property is used to set the module identifier.

        Parameters
        ----------
        source : SwanStorage
            Swan module (.swant)

        Returns
        -------
        TestModule
            Instance of TestModule.
        """
        # source.check_swan_version()
        result = self._parse(Reader.parse_body, source)
        return moduleOfAst(source.name, result)

    def module_interface(self, source: SwanStorage) -> S.ModuleInterface:
        """Parse a Swan interface from a SwanStorage object.

            The *content()* method is called to get the code.

            The *name* property is used to set the module identifier.

        Parameters
        ----------
        source : SwanStorage
            Swan interface (.swani)

        Returns
        -------
        ModuleInterface
            Instance of ModuleInterface.
        """
        source.check_swan_version()
        result = self._parse(Reader.parse_interface, source)
        return interfaceOfAst(source.name, result.Item1)

    def declaration(self, source: SwanStorage) -> S.Declaration:
        """Parse a Swan declaration:
          type, const, sensor, group, use, operator (signature or with body).

        Parameters
        ----------
        source : SwanStorage
            Single Swan declaration

        Returns
        -------
        Declaration
            Corresponding declaration object
        """
        ast = self._parse(Reader.parse_declaration, source)
        return declarationOfAst(ast)

    def equation(self, source: SwanStorage) -> S.Equation:
        """Parse a Swan equation.

        Parameters
        ----------
        source : SwanStorage
            Swan equation text

        Returns
        -------
        Equation
            Corresponding Equation object
        """
        ast = self._parse(Reader.parse_equation, source)
        return equationOfAst(ast)

    def expression(self, source: SwanStorage) -> S.expressions:
        """Parse a Swan expression

        Parameters
        ----------
        source : SwanStorage
            Swan expression text

        Returns
        -------
        Expression
            Corresponding expression object
        """
        ast = self._parse(Reader.parse_expr, source)
        return expressionOfAst(ast)

    def scope_section(self, source: SwanStorage) -> S.ScopeSection:
        """Parse a Swan scope section

        Parameters
        ----------
        source : str
            Swan scope section text

        Returns
        -------
        ScopeSection
            Corresponding scope section object
        """
        ast = self._parse(Reader.parse_scope_section, source)
        return scopeSectionOfAst(ast)

    def op_expr(self, source: SwanStorage) -> S.OperatorExpression:
        """Parse a Swan operator expression

        Parameters
        ----------
        source : SwanStorage
            Swan code for operator expression

        Returns
        -------
        OperatorExpression
            Instance of the operator expression object
        """
        ast = self._parse(Reader.parse_op_expr, source)
        return operatorExprOfAst(ast)

    def operator_block(self, source: SwanStorage) -> Union[S.OperatorBase, S.OperatorExpression]:
        """Parse a Swan operator block

        *operator_block* ::= *operator* | *op_expr*

        Parameters
        ----------
        source : SwanStorage
            Swan code for operator block

        Returns
        -------
        Union[S.OperatorBase, S.OperatorExpression]
            Instance of the *operator* or *op_expr*
        """
        ast = self._parse(Reader.parse_operator_block, source)
        return operatorBlockOfAst(ast)

    def operator_decl(self, source: SwanStorage) -> Union[S.Operator, S.Signature, None]:
        """Parse a Swan operator declaration

        Parameters
        ----------
        source : SwanStorage
            Swan operator text or operator interface

        Returns
        -------
        S.Operator|S.Signature
            Instance of the operator, or its signature
            Returns None if expected parsing error occurs (for markup parsing)
        """
        ast = self._parse(Reader.parse_user_operator, source, parse_error_ok=True)
        if ast is None:
            return None
        if ast.OpBody.IsSDEmpty:
            return signatureOfAst(ast)
        return operatorOfAst(ast)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List, Union, cast

from ansys.scadeone.core import project  # noqa: F401
from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.storage import SwanFile
import ansys.scadeone.core.swan as swan

from .loader import SwanParser


class Model:
    """Model handling class.
    A model contains module and interface declarations.

    Loading of Swan sources is lazy.
    """

    def __init__(self) -> None:
        self._project = None
        self._parser = None
        # dictionaries of Swan module, interface and test harness
        # Key is the Swan name of the module, interface or test harness
        self._bodies = {}  # _bodies["N1::N2::M"] = ModuleBody | SwanFile
        self._interfaces = {}
        self._harnesses = {}

    def _add_module(self, swan_elt: Union[swan.SwanFile, swan.Module], where: dict) -> None:
        # add element to the corresponding dictionary where'
        where[Model._get_swan_name(swan_elt)] = swan_elt
        if isinstance(swan_elt, swan.Module):
            swan_elt.owner = self

    def add_body(self, swan_elt: Union[swan.SwanFile, swan.ModuleBody]) -> None:
        """Add a module body to the model.

        The module body can be a SwanFile or a ModuleBody object. If it is a SwanFile,
        it is not loaded: this happens when needed. If it is a ModuleBody it is directly added
        to the model.


        Parameters
        ----------
        swan_elt : Union[swan.SwanFile, swan.ModuleBody]
            Content of the module body
        """
        self._add_module(swan_elt, self._bodies)

    def add_interface(self, swan_elt: Union[swan.SwanFile, swan.ModuleInterface]) -> None:
        """Add a module interface to the model.

        The module interface can be a SwanFile or a ModuleInterface object. If it is a SwanFile,
        it is not loaded: this happens when needed. If it is a ModuleInterfacee it is directly added
        to the model.


        Parameters
        ----------
        swan_elt : Union[swan.SwanFile, swan.ModuleInterface]
            Content of the module interface
        """
        self._add_module(swan_elt, self._interfaces)

    def add_harness(self, swan_elt: Union[swan.SwanFile, swan.TestModule]) -> None:
        self._add_module(swan_elt, self._harnesses)

    def module_exists(self, module: Union[swan.SwanFile, swan.Module]) -> bool:
        """Check if a module exists in the model.

        Parameters
        ----------
        module :Union[swan.SwanFile, swan.Module]
            Module source or module object to check. In case of a source, it is not loaded.

        Returns
        -------
        bool
            True if the module exists, False otherwise.
        """
        where = None
        if isinstance(module, swan.Module):
            module_name = cast(swan.Module, module).name.as_string
            if isinstance(module, swan.ModuleBody):
                where = self._bodies
            elif isinstance(module, swan.ModuleInterface):
                where = self._interfaces
            elif isinstance(module, swan.TestModule):
                where = self._harnesses
        else:
            module_name = Model._get_swan_name(module)
            if cast(SwanFile, module).is_module:
                where = self._bodies
            elif cast(SwanFile, module).is_interface:
                where = self._interfaces
            elif cast(SwanFile, module).is_test:
                where = self._harnesses
        if where:
            return module_name in where.keys()
        return False

    @staticmethod
    def _get_swan_name(swan_elt: Union[swan.SwanFile, swan.Module]) -> str:
        """Returns the name of a Swan element.

        Parameters
        ----------
        swan_elt : Union[swan.SwanFile, swan.TestModule]
            Swan element

        Returns
        -------
        str
            Name of the Swan element.
        """
        if isinstance(swan_elt, SwanFile):
            return swan.Module.module_name_from_path(swan_elt.path)
        return swan_elt.name.as_string.replace("-", "::")

    def configure(self, project_instance: "project.IProject"):
        """Configures model for a given project. The configuration
        associates the project and the model and prepares internal data to
        store module bodies and interfaces.

        It is called by :py:attr:`ansys.scadeone.core.project.Project.model`."""

        if project_instance.storage.exists():
            for swan_file in project_instance.swan_sources(all=True):
                module_name = swan.Module.module_name_from_path(swan_file.path)
                if swan_file.is_module:
                    self._bodies[module_name] = swan_file
                elif swan_file.is_interface:
                    self._interfaces[module_name] = swan_file
                # TODO use this once test harness visitor and creator are implemented
                # elif swan_file.is_test:
                #     self._harnesses[module_name] = swan_file

        self._project = project_instance
        self._parser = SwanParser(self.project.app.logger)
        return self

    @property
    def project(self) -> "project.IProject":
        """Model project, as a Project object."""
        return self._project

    @property
    def parser(self) -> SwanParser:
        """Swan parser."""
        return self._parser

    def _load_source(
        self,
        swan_f: Union[SwanFile, swan.ModuleBody, swan.ModuleInterface, swan.TestModule],
    ) -> swan.Module:
        """Read a Swan file (.swan or .swani or .swant)

        Parameters
        ----------
        swan_f : Union[SwanFile, swan.ModuleBody, swan.ModuleInterface, swan.TestModule]
            Swan source code or module

        Returns
        -------
        Module
            Swan Module, either a ModuleBody, a ModuleInterface or a TestModule.

        Raises
        ------
        ScadeOneException
            - Error when file has not the proper suffix
            - Parse error
        """
        if (
            isinstance(swan_f, swan.ModuleBody)
            or isinstance(swan_f, swan.ModuleInterface)
            or isinstance(swan_f, swan.TestModule)
        ):
            return swan_f
        elif swan_f.is_module:
            ast = self.parser.module_body(swan_f)
            ast.source = str(swan_f.path)
        elif swan_f.is_interface:
            ast = self.parser.module_interface(swan_f)
            ast.source = str(swan_f.path)
        elif swan_f.is_test:
            ast = self.parser.test_harness(swan_f)
        else:
            raise ScadeOneException(f"Model.load_source: unexpected file kind {swan_f.path}.")
        ast.owner = self
        return ast

    @property
    def all_modules_loaded(self) -> bool:
        """Returns True when all Swan modules have been loaded."""
        for module in self.all_modules:
            if not (
                isinstance(module, swan.ModuleBody)
                or isinstance(module, swan.ModuleInterface)
                or isinstance(module, swan.TestModule)
            ):
                return False
        return True

    @property
    def modules(self) -> List[swan.Module]:
        """Returns all *loaded* Module objects (module body, interface, test module) as a list."""
        modules_list = [mod for mod in self.all_modules if isinstance(mod, swan.Module)]
        return modules_list

    @property
    def all_modules(self) -> List[Union[swan.Module, SwanFile]]:
        """Returns *all* modules (loaded Module objects, or not-yet-loaded module files) as a list."""
        modules = (
            list(self._bodies.values())
            + list(self._interfaces.values())
            + list(self._harnesses.values())
        )
        return modules

    def get_module_body(self, name: str) -> Union[swan.ModuleBody, None]:
        """Returns module body of name 'name'"""
        if name not in self._bodies:
            return None
        if isinstance(self._bodies.get(name), SwanFile):
            self._bodies[name] = self._load_source(self._bodies[name])
        return self._bodies[name]

    def get_module_interface(self, name: str) -> Union[swan.ModuleInterface, None]:
        """Returns module interface of name 'name'"""
        if name not in self._interfaces:
            return None
        if isinstance(self._interfaces.get(name), SwanFile):
            self._interfaces[name] = self._load_source(self._interfaces[name])
        return self._interfaces[name]

    def get_test_harness(self, name: str) -> Union[swan.TestModule, None]:
        """Returns module test harness of name 'name'"""
        if name not in self._harnesses:
            return None
        if isinstance(self._harnesses.get(name), SwanFile):
            self._harnesses[name] = self._load_source(self._harnesses[name])
        return self._harnesses[name]

    def get_module_from_pathid(self, pathid: str, module: swan.Module) -> Union[swan.Module, None]:
        """Return the :py:class:`Module` instance for a given *pathid*
        A *path* is of the form *[ID ::]+ ID*, where the last ID is the object
        name, and the "ID::ID...::" is the module path.

        If the *pathid* has no path part (reduced to ID), return *module*.

        Parameters
        ----------
        pathid : str
            object full path

        module : Module
            Context module where the search occurs.

        Returns
        -------
        Union[S.Module, None]
            Module of the object, or None if not module found
        """
        ids = pathid.split("::")

        if len(ids) == 1:
            return module

        if len(ids) == 2:
            # case M::ID
            if module.name.as_string == ids[0]:
                # case M::ID inside M (can happen from a search).
                # No use directives in that case
                return module
            use = module.get_use_directive(ids[0])
            if not use:
                # if not in module, try in interface.
                # not: module can be an interface already, its interface is None
                interface = module.interface()
                if not interface:
                    return None
                use = interface.get_use_directive(ids[0])
                if not use:
                    return None
            module_path = cast(swan.UseDirective, use).path.as_string
        else:
            module_path = "::".join(ids[0:-1])
        m = self.get_module_body(module_path)
        if m is None:
            m = self.get_module_interface(module_path)
        return m

    @staticmethod
    def get_path_in_module(
        declaration: swan.Declaration, module: swan.Module
    ) -> swan.PathIdentifier:
        """Get the path of declaration in module. If the declaration is in module, return the identifier,
        else return the path of the declaration in the module based on the use directives.

        Parameters
        ----------
        item: SwanItem
            Declaration to get the path for.
        module: Module
            Module where the declaration is used.

        Returns
        -------
        PathIdentifier
            Path of the item in the module.

        Raises
        ------
        ScadeOneException
            - If the declaration does not have a module.
            - If the module does not exist.
            - If the module does not have a use directive for the declaration module.
        """

        if declaration.module is None:
            raise ScadeOneException(f"{declaration.get_full_path()} does not have a module.")
        if module is None:
            raise ScadeOneException("Module does not exist.")

        if cast(swan.Module, declaration.module).name == module.name:
            # The operator is in the same module as the instance
            return swan.PathIdentifier([declaration.id])

        # `declaration`` is in another module

        # Get the use directive of the operator module including body and/or interface
        module_uses = module.use_directives
        if module.model:
            # if model is defined (required by called functions), get the peer module from the model
            peer_module = (
                module.interface() if isinstance(module, swan.ModuleBody) else module.body()
            )
        else:
            peer_module = None
        peer_uses = cast(swan.Module, peer_module).use_directives if peer_module else []

        # Find use directive for the declaration module in all uses
        decl_module_path = cast(swan.Module, declaration.module).name.as_string

        uses = [
            use
            for use_list in [module_uses, peer_uses]
            for use in use_list
            if use.path.as_string == decl_module_path
        ]
        if not uses:
            raise ScadeOneException(
                f"Missing use directive in module {module.name.as_string} for item {declaration.get_full_path()}."
            )
        # Get the module part of the use directive
        # If the use directive has an alias, use it. Otherwise, use the last part of the path.
        mod_part = (
            uses[0].alias.value
            if uses[0].alias
            else swan.PathIdentifier.split(uses[0].path.as_string)[-1]
        )
        return swan.PathIdentifier.from_string(f"{mod_part}::{declaration.id}")

    @property
    def types(self) -> List[swan.TypeDecl]:
        """Returns a list of type declarations."""
        types = []
        for decls in self.filter_declarations(lambda x: isinstance(x, swan.TypeDeclarations)):
            for type in cast(swan.TypeDeclarations, decls).types:
                types.append(type)
        return types

    @property
    def sensors(self) -> List[swan.SensorDecl]:
        """Returns a list of sensor declarations."""
        sensors = []
        for decls in self.filter_declarations(lambda x: isinstance(x, swan.SensorDeclarations)):
            for sensor in cast(swan.SensorDeclarations, decls).sensors:
                sensors.append(sensor)
        return sensors

    @property
    def constants(self) -> List[swan.ConstDecl]:
        """Returns a list of constant declarations."""
        consts = []
        for decls in self.filter_declarations(lambda x: isinstance(x, swan.ConstDeclarations)):
            for const in cast(swan.ConstDeclarations, decls).constants:
                consts.append(const)
        return consts

    @property
    def groups(self) -> List[swan.GroupDecl]:
        """Returns a list of group declarations."""
        groups = []
        for decls in self.filter_declarations(lambda x: isinstance(x, swan.GroupDeclarations)):
            for group in cast(swan.GroupDeclarations, decls).groups:
                groups.append(group)
        return groups

    @property
    def operators(self) -> List[swan.Operator]:
        """Returns a list of operator declarations."""
        return [
            cast(swan.Operator, operator)
            for operator in self.filter_declarations(lambda x: isinstance(x, swan.Operator))
        ]

    @property
    def signatures(self) -> List[swan.Signature]:
        """Returns a list of operator signature declarations."""
        return [
            cast(swan.Signature, signature)
            for signature in self.filter_declarations(
                lambda x: isinstance(x, swan.Signature) and not isinstance(x, swan.Operator)
            )
        ]

    def _load_module(self, name: str, mod_dict: dict) -> None:
        swan_file = mod_dict.get(name)
        if swan_file and isinstance(swan_file, SwanFile):
            mod_dict[name] = self._load_source(swan_file)

    def load_module_body(self, name: str) -> None:
        self._load_module(name, self._bodies)

    def load_module_interface(self, name: str) -> None:
        self._load_module(name, self._interfaces)

    def load_test_harness(self, name: str) -> None:
        self._load_module(name, self._harnesses)

    def load_all_modules(
        self, bodies: bool = True, interfaces: bool = True, harnesses: bool = True
    ) -> None:
        """Loads systematically all modules.

        Parameters
        ----------
        bodies : bool, optional
            Includes module bodies
        interfaces : bool, optional
            Includes module interfaces
        harnesses : bool, optional
            Includes test harnesses
        """
        for cond, data, load_fn in [
            (bodies, self._bodies, self.load_module_body),
            (interfaces, self._interfaces, self.load_module_interface),
            (harnesses, self._harnesses, self.load_test_harness),
        ]:
            if cond:
                for name in data.keys():
                    load_fn(name)

    @property
    def declarations(self) -> List[swan.GlobalDeclaration]:
        """Declarations found in all modules/interfaces as a list.

        The Swan code of a module/interface is loaded if not yet loaded.
        """

        declarations = []
        for data, load_fn in [
            (self._interfaces, self.load_module_interface),
            (self._bodies, self.load_module_body),
            (self._harnesses, self.load_test_harness),
        ]:
            for swan_code, swan_object in data.items():
                if isinstance(swan_object, SwanFile):
                    swan_object = self._load_source(swan_object)
                    data[swan_code] = swan_object
                elif swan_object is None:
                    swan_object = load_fn(swan_code)
                    data[swan_code] = swan_object
                for decl in swan_object.declarations:
                    declarations.append(cast(swan.GlobalDeclaration, decl))
        return declarations

    def filter_declarations(self, filter_fn) -> List[swan.GlobalDeclaration]:
        """Returns declarations matched by a filter.

        Parameters
        ----------
        filter_fn : function
            A function of one argument of type S.GlobalDeclaration, returning True or False.

        Returns
        -------
        List[S.GlobalDeclaration]
            List of matching declarations.
        """
        return list(filter(filter_fn, self.declarations))

    def find_declaration(self, predicate_fn) -> Union[swan.GlobalDeclaration, None]:
        """Finds a declaration for which predicate_fn returns True.

        Parameters
        ----------
        predicate_fn : function
            Function taking one S.GlobalDeclaration as argument and
            returning True when some property holds, else False.

        Returns
        -------
        Union[S.GlobalDeclaration, None]
            Found declaration or None.
        """
        for decl in self.filter_declarations(predicate_fn):
            return decl
        return None



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The **storage** module contains classes that abstract a storage container.
Currently, containers can be a file or a string, or other data

The *content()* method returns the storage content.

The *source* property gives the origin of the storage,
Source is used for error messages.
"""

from abc import ABC, abstractmethod
import json
from pathlib import Path
import re
from typing import Optional, Union

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.versioning import FormatVersions


class Storage(ABC):
    """Top-level class for storage: Project, Swan code, etc.

    Storage abstracts the data persistence. It has a _source_ property
    which gives the origin of the data (file name, string, etc.)

    The *content()* method is responsible for returning the data contained
    by the source. For now, content is a string, its interpretation is
    made by its consumer.
    """

    def __init__(self, source: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self._source = source

    @property
    def source(self) -> str:
        """Source origin: file name, string, etc."""
        return self._source

    @source.setter
    def source(self, source: str) -> None:
        """Change the source"""
        self._source = source

    @abstractmethod
    def exists(self) -> bool:
        """True when source exists."""
        pass

    @abstractmethod
    def content(self) -> str:
        """Content of the source."""
        pass

    @abstractmethod
    def set_content(self, data: str) -> str:
        """Sets content of the source."""
        pass


class FileStorage(Storage):
    """Base class for storage as a file.
    file is stored as absolute, posix-style.
    """

    def __init__(self, file: Union[str, Path], **kwargs) -> None:
        path = Path(file) if isinstance(file, str) else file
        path = path.resolve().as_posix()
        super().__init__(source=path)

    @property
    def path(self) -> Path:
        """Saved path."""
        return Path(self.source)

    def exists(self) -> bool:
        """True when file exists."""
        return self.path.exists()

    def content(self) -> str:
        """Content of file."""
        if self.exists():
            content = self.path.read_text(encoding="utf-8")
            return content
        raise ScadeOneException(f"FileStorage.content(): no such file: {self.path}.")

    def set_content(self, data: str) -> str:
        """Sets content and write it to underlying file."""
        self.path.write_text(data)


class StringStorage(Storage):
    """Base class for storage provided as a string."""

    def __init__(self, text: str, **kwargs) -> None:
        super().__init__(source="<string>")
        self._text = text

    def exists(self) -> bool:
        """Always return True."""
        return True

    def content(self) -> str:
        """Content of string."""
        return self._text

    def set_content(self, data: str) -> str:
        """Set content of the file."""
        self._text = data


class JSONStorage(object):
    """Toplevel class for JSON-related storage."""

    def __init__(self, asset: Storage, **kwargs) -> None:
        super().__init__(**kwargs)
        self._asset = asset
        self._json = None

    @property
    def json(self):
        """JSON content.
        Any modification is propagated to the underlying JSON object."""
        return self._json

    @json.setter
    def json(self, json_data):
        """Update JSON content"""
        self._json = json_data

    def load(self, **kw):
        """Loads content of JSON data into json property and returns `self`.

        See `json.loads() <https://docs.python.org/3/library/json.html>`_
        for detailed interface.
        """
        self.json = json.loads(self._asset.content(), **kw)
        return self

    def dump(self, **kw):
        """Uses `self.json` to update storage content and returns self.

        See `json.dumps() <https://docs.python.org/3/library/json.html>`_
        for detailed interface.
        """
        data = json.dumps(self.json, **kw)
        self._asset.set_content(data)
        return self

    def exists(self) -> bool:
        """True when a source exists."""
        pass


# Swan related storage
# ====================


class SwanStorage(ABC):
    """Toplevel class for Swan input code."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        """Name for module or interface."""
        pass

    @property
    def version(self) -> Union[dict, None]:
        """Swan version."""
        return None

    @staticmethod
    def extract_version(source: str) -> Union[dict, None]:
        """Extracts version information from a Swan source.

        Parameters
        ----------
        source : str
            Version information as a string.

        Returns
        -------
        Union[dict, None]
            Either the version information as a dictionary, or None if no version found.
        """
        m = re.match(r"^--\s*version\s+(?P<ver>.*)$", source, re.MULTILINE)
        if m:
            infos = {
                k: v
                for k, v in re.findall(r"(\w+):\s*(\d+\.\d+)", m["ver"])
                if k in FormatVersions.formats
            }
            infos["version"] = m[0]
            return infos
        return None

    def check_swan_version(self):
        """Check Swan version information.

        Raises
        ------
        ScadeOneException
            When version information is missing or invalid.
        """
        version = self.version
        if version is None:
            raise ScadeOneException("No version information found.")
        for k in ("swan", "graph"):
            if k not in version:
                raise ScadeOneException(f"Missing version information for {k}.")
            FormatVersions.check(k, version[k])
        return version

    def check_swant_version(self):
        """Check Swan test harness version information.

        Raises
        ------
        ScadeOneException
            When version information is missing or invalid.
        """
        version = self.check_swan_version()
        if "swant" not in version:
            raise ScadeOneException("Missing version information for swant.")
        FormatVersions.check("swant", version["swant"])
        return version


class SwanFile(FileStorage, SwanStorage):
    """Swan code within a file.

    Parameters
    ----------
    file : Path
        File containing the Swan source."""

    def __init__(self, file: Union[str, Path]) -> None:
        super().__init__(file=file)

    @property
    def name(self) -> str:
        """Returns basename of source file (no suffix)."""
        return self.path.stem

    @property
    def is_module(self) -> bool:
        """True when file is a module code."""
        return self.path.suffix == ".swan"

    @property
    def is_interface(self) -> bool:
        """True when file is an interface code."""
        return self.path.suffix == ".swani"

    @property
    def is_test(self) -> bool:
        """True when file is a test code."""
        return self.path.suffix == ".swant"

    @property
    def version(self) -> Union[dict, None]:
        """Swan version information."""
        try:
            with self.path.open() as fd:
                return self.extract_version(fd.readline())
        except:  # noqa: E722
            return None


class SwanString(StringStorage, SwanStorage):
    """Swan code within a string

    Parameters
    ----------
    swan_code : str
        String containing the Swan source.

    name: str [optional]
        Name that can be used for a module or interface identifier.
    """

    def __init__(self, swan_code: str, name: Optional[str] = None) -> None:
        super().__init__(text=swan_code)
        self._name = name if name else "from_string"

    @property
    def name(self) -> str:
        """Name attribute."""
        return self._name

    @property
    def source(self) -> str:
        """Source string."""
        return f"string:<{self.content()}>"

    @property
    def version(self) -> Union[dict, None]:
        """Swan version."""
        return self.extract_version(self.content())


# Project Storage
# -------------


class ProjectStorage(JSONStorage):
    """Base class for project storage."""

    def __init__(self, **kwargs) -> None:
        super().__init__(asset=self, **kwargs)


class ProjectFile(FileStorage, ProjectStorage):
    """Project as a file."""

    def __init__(self, file: Union[str, Path]) -> None:
        super().__init__(file=file)


# Job Storage
# --------


class JobStorage(JSONStorage):
    """Base class for job asset."""

    def __init__(self, **kwargs) -> None:
        super().__init__(asset=self, **kwargs)


class JobFile(FileStorage, JobStorage):
    """Job asset as a file."""

    def __init__(self, file: Union[str, Path]) -> None:
        super().__init__(file=file)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import platform
from pathlib import Path
from typing import List, Optional, Union

from ansys.scadeone.core.common.exception import ScadeOneException
from ansys.scadeone.core.common.logger import LOGGER, ScadeOneLogger
from ansys.scadeone.core.common.storage import ProjectFile, ProjectStorage
from ansys.scadeone.core.interfaces import IScadeOne as IScadeOne
from ansys.scadeone.core.project import Project
from ansys.scadeone.core.model.model import Model


class ScadeOne(IScadeOne):
    """Scade One API."""

    def __init__(self, install_dir: Optional[Union[str, Path]] = None) -> None:
        self._logger = LOGGER
        self._projects = []
        self._install_dir = Path(install_dir) if isinstance(install_dir, str) else install_dir
        self._model = Model()
        self._tools = self._Tools(self._install_dir)

    class _Tools:
        """Private class for Scade One installation tools."""

        def __init__(self, install_dir: Optional[Path] = None) -> None:
            self._install_dir = install_dir

        @property
        def job_launcher(self) -> Union[Path, None]:
            """Get job launcher executable path if exists."""
            if not self._install_dir:
                return None
            launcher = (
                self._install_dir / "tools/job_launcher/scade_one_job_launcher.exe"
                if platform.system() == "Windows"
                else self._install_dir / "tools/job_launcher/scade_one_job_launcher"
            )
            if launcher.exists():
                return launcher
            return None

    @property
    def model(self) -> Model:
        return self._model

    @property
    def projects(self) -> List[Project]:
        """Returns the loaded projects.

        Returns
        -------
        List[Project]
           Loaded projects.
        """
        return self._projects

    @property
    def install_dir(self) -> Union[Path, None]:
        """Installation directory as given when creating the ScadeOne instance."""
        return self._install_dir

    @property
    def version(self) -> str:
        """API version."""
        from ansys.scadeone.core import __version__

        return __version__

    @property
    def logger(self) -> ScadeOneLogger:
        return self._logger

    @property
    def tools(self) -> _Tools:
        return self._tools

    # For context management
    def __enter__(self) -> "ScadeOne":
        self.logger.info("Entering context")
        return self

    def __exit__(self, exc_type, exc_value, exc_tb) -> bool:
        if exc_type and not isinstance(exc_type, Union[ScadeOneException, SystemExit]):
            msg = f"Exiting on exception {exc_type}"
            if exc_value:
                msg += f" with value {exc_value}"
            self.logger.exception(msg)
        self.close()
        # propagate exception
        return False

    # end context management

    def close(self) -> None:
        """Close application, releasing any connection."""
        pass

    def __del__(self) -> None:
        self.close()

    def load_project(self, storage: Union[ProjectStorage, str, Path]) -> Union[Project, None]:
        """Load a Scade One project.

        Parameters
        ----------
        storage : Union[ProjectStorage, str, Path]
            Storage containing project data.

        Returns
        -------
        Project|None
            Project object, or None if file does not exist.
        """
        if not isinstance(storage, ProjectStorage):
            storage = ProjectFile(storage)
        if not storage.exists():
            self.logger.error(f"Project does not exist {storage.source}")
            return None
        project = Project(self, storage)
        self._projects.append(project)
        self._model.configure(project)
        return project

    def subst_in_path(self, path: str) -> str:
        """Substitute $(SCADE_ONE_LIBRARIES_DIR) in path.

        if :py:attr:`ScadeOne.install_dir` is None, no change is made.
        """
        if self.install_dir:
            return path.replace("$(SCADE_ONE_LIBRARIES_DIR)", str(self.install_dir / "libraries"))

        return path

    def new_project(self, storage: Union[str, Path]) -> Optional[Project]:
        """Create a new project.

        Parameters
        ----------
        storage : Union[str, Path]
            Project location.

        Returns
        -------
        Project
            Project object.
        """

        if not storage:
            raise ScadeOneException("Storage must be provided.")

        if isinstance(storage, str):
            storage = Path(storage)

        if storage.exists():
            raise ScadeOneException(f"'{storage.name}' project already exists.")

        storage = ProjectFile(storage)

        from ansys.scadeone.core.svc.swan_creator.project_creator import (
            ProjectAdder,
            ProjectFactory,
        )

        project = ProjectFactory.create_project(self, storage)
        ProjectAdder.add_project(self, project)
        return project



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
The :py:class:`ScadeOneLogger` class is a subclass of the `logging.Logger` class.
As a singleton, the `ScadeOneLogger` class is initialized once and used
throughout the application using the `logger.LOGGER` object or the `logger`
application attribute.

The default logger is set to log to a file named `pyscadeone.log` and to
the console. The log level is set to `DEBUG`.

One can set the logger to a different logger by setting the `logger` attribute
as in the following example:

.. code:: python

    LOGGER.logger = logging.getLogger("MyLogger")
"""  # numpydoc ignore

# cSpell:ignore levelname

from pathlib import Path
import logging

# From: https://docs.python.org/3/howto/logging-cookbook.html#logging-cookbook


class ScadeOneLogger:
    """Class handling the singleton logger."""

    _Logger = None

    def __init__(self) -> None: ...

    @property
    def logger(self) -> logging.Logger:
        """Return the logger instance. The internal logger can be set
        by setting the `logger` attribute."""
        if ScadeOneLogger._Logger is None:
            self._init_logger()
        return self._Logger

    @logger.setter
    def logger(self, logger: logging.Logger) -> None:
        """Set the logger instance."""
        # Unset PyScadeOne default logger if it exists
        pyscadeone_logger = logging.getLogger("PyScadeOne")
        for handler in pyscadeone_logger.handlers:
            if isinstance(handler, logging.FileHandler):
                handler.close()
                Path(handler.baseFilename).unlink(missing_ok=True)
            pyscadeone_logger.removeHandler(handler)
        ScadeOneLogger._Logger = logger

    def _init_logger(self):
        """Set the logger instance.

        If the logger instance is not set, create a file logger.
        """
        if ScadeOneLogger._Logger is not None:
            return
        logger = logging.getLogger("PyScadeOne")
        logger.setLevel(logging.DEBUG)

        # create formatter and add it to the handlers
        format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        formatter = logging.Formatter(format)
        # create file handler which logs even debug messages
        fh = logging.FileHandler("pyscadeone.log")
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        # create console handler with a higher log level
        ch = logging.StreamHandler()
        ch.setLevel(logging.ERROR)
        ch.setFormatter(formatter)
        # add the handlers to the logger
        logger.addHandler(ch)
        ScadeOneLogger._Logger = logger

    def debug(self, msg: str, *args, **kwargs) -> None:
        """Log a message with severity DEBUG on the logger."""
        self._init_logger()
        self._Logger.debug(msg, *args, **kwargs)

    def info(self, msg: str, *args, **kwargs) -> None:
        """Log a message with severity INFO on the logger."""
        self._init_logger()
        self._Logger.info(msg, *args, **kwargs)

    def warning(self, msg: str, *args, **kwargs) -> None:
        """Log a message with severity WARNING on the logger."""
        self._init_logger()
        self._Logger.warning(msg, *args, **kwargs)

    def exception(self, msg: str, *args, **kwargs) -> None:
        """Log an exception message with severity ERROR on the logger."""
        self._init_logger()
        self._Logger.exception(msg, *args, **kwargs)

    def error(self, msg: str, *args, **kwargs) -> None:
        """Log an error message with severity ERROR on the logger."""
        self._init_logger()
        self._Logger.error(msg, *args, **kwargs)

    def critical(self, msg: str, *args, **kwargs) -> None:
        """Log a message with severity CRITICAL on the logger."""
        self._init_logger()
        self._Logger.critical(msg, *args, **kwargs)


LOGGER = ScadeOneLogger()
